from pathlib import Path

import pytest

from ngio import Image, create_empty_ome_zarr, open_image
from ngio.transforms import ZoomTransform
from ngio.utils import NgioValueError


@pytest.mark.parametrize(
    "zarr_name",
    [
        "test_image_yx.zarr",
        "test_image_cyx.zarr",
        "test_image_zyx.zarr",
        "test_image_czyx.zarr",
        "test_image_c1yx.zarr",
        "test_image_tyx.zarr",
        "test_image_tcyx.zarr",
        "test_image_tzyx.zarr",
        "test_image_tczyx.zarr",
    ],
)
def test_open_image(images_v04: dict[str, Path], zarr_name: str):
    path = images_v04[zarr_name]
    image = open_image(path)
    assert isinstance(image, Image)


@pytest.mark.parametrize(
    ("shape1", "shape2", "axes1", "axes2", "should_raise"),
    [
        ((1, 3, 10, 10), (1, 3, 12, 10), "czyx", "czyx", False),
        ((1, 10, 10), (1, 1, 10, 10), "cyx", "czyx", True),
        ((3, 10, 10), (3, 3, 10, 10), "zyx", "czyx", False),
        ((1, 1, 10, 10), (1, 1, 10, 10), "tcyx", "czyx", True),
    ],
)
def test_image_require_axes_match(
    tmp_path: Path, shape1, shape2, axes1, axes2, should_raise: bool
):
    path1 = tmp_path / "image1.zarr"
    path2 = tmp_path / "image2.zarr"
    create_empty_ome_zarr(
        store=path1, shape=shape1, axes_names=axes1, xy_pixelsize=0.5, levels=1
    )
    create_empty_ome_zarr(
        store=path2, shape=shape2, axes_names=axes2, xy_pixelsize=0.5, levels=1
    )
    img1 = open_image(path1)
    img2 = open_image(path2)

    if should_raise:
        with pytest.raises(NgioValueError):
            img1.require_axes_match(img2)
    else:
        img1.require_axes_match(img2)


@pytest.mark.parametrize(
    ("shape1", "shape2", "axes1", "axes2", "allow_singleton", "should_raise"),
    [
        ((1, 3, 10, 10), (1, 3, 10, 10), "czyx", "czyx", False, False),
        ((1, 3, 10, 10), (1, 3, 10, 10), "czyx", "czyx", True, False),
        ((1, 1, 10, 10), (1, 3, 10, 10), "czyx", "czyx", True, False),
        ((3, 10, 10), (3, 3, 10, 10), "zyx", "czyx", True, False),
        ((1, 1, 10, 10), (1, 3, 10, 10), "czyx", "czyx", False, True),
        ((1, 2, 10, 10), (1, 3, 10, 10), "czyx", "czyx", False, True),
    ],
)
def test_image_require_dimensions_match(
    tmp_path: Path, shape1, shape2, axes1, axes2, allow_singleton, should_raise: bool
):
    path1 = tmp_path / "image1.zarr"
    path2 = tmp_path / "image2.zarr"
    create_empty_ome_zarr(
        store=path1, shape=shape1, axes_names=axes1, xy_pixelsize=0.5, levels=1
    )
    create_empty_ome_zarr(
        store=path2, shape=shape2, axes_names=axes2, xy_pixelsize=0.5, levels=1
    )
    img1 = open_image(path1)
    img2 = open_image(path2)

    if should_raise:
        with pytest.raises(NgioValueError):
            img1.require_dimensions_match(img2, allow_singleton=allow_singleton)
    else:
        img1.require_dimensions_match(img2, allow_singleton=allow_singleton)


@pytest.mark.parametrize(
    ("shape1", "shape2", "axes1", "axes2", "xy_pixelsize", "should_raise"),
    [
        ((1, 3, 10, 10), (1, 3, 10, 10), "czyx", "czyx", 1.0, False),
        ((1, 3, 10, 10), (1, 3, 20, 20), "czyx", "czyx", 0.5, False),
        ((1, 3, 10, 10), (1, 3, 20, 20), "czyx", "czyx", 2.0, True),
    ],
)
def test_image_require_can_be_rescaled(
    tmp_path: Path,
    shape1,
    shape2,
    axes1,
    axes2,
    xy_pixelsize: float,
    should_raise: bool,
):
    path1 = tmp_path / "image1.zarr"
    path2 = tmp_path / "image2.zarr"
    create_empty_ome_zarr(
        store=path1, shape=shape1, axes_names=axes1, xy_pixelsize=1.0, levels=1
    )
    create_empty_ome_zarr(
        store=path2, shape=shape2, axes_names=axes2, xy_pixelsize=xy_pixelsize, levels=1
    )
    img1 = open_image(path1)
    img2 = open_image(path2)

    # Also test with transforms
    if should_raise:
        with pytest.raises(NgioValueError):
            img1.require_rescalable(img2)
        return None

    img1.require_rescalable(img2)

    # Also test with transforms
    zoom = ZoomTransform(
        input_image=img2,
        target_image=img1,
        order="nearest",
    )
    img1_data = img1.get_as_numpy()
    img2_data = img2.get_as_numpy(transforms=[zoom])
    assert img1_data.shape == img2_data.shape


@pytest.mark.parametrize(
    ("shape1", "shape2", "axes1", "axes2", "xy_pixelsize", "should_raise"),
    [
        ((111, 111), (112, 112), "yx", "yx", 0.99, False),
        ((111, 111), (111, 111), "yx", "yx", 2.0, True),
        ((111, 111), (55, 55), "yx", "yx", 2.0, False),
        ((111, 111), (56, 56), "yx", "yx", 2.0, False),
        ((111, 111), (54, 54), "yx", "yx", 2.0, True),
    ],
)
def test_image_require_can_be_rescaled2(
    tmp_path: Path,
    shape1,
    shape2,
    axes1,
    axes2,
    xy_pixelsize: float,
    should_raise: bool,
):
    path1 = tmp_path / "image1.zarr"
    path2 = tmp_path / "image2.zarr"
    create_empty_ome_zarr(
        store=path1, shape=shape1, axes_names=axes1, xy_pixelsize=1.0, levels=1
    )
    create_empty_ome_zarr(
        store=path2, shape=shape2, axes_names=axes2, xy_pixelsize=xy_pixelsize, levels=1
    )
    img1 = open_image(path1)
    img2 = open_image(path2)

    # Also test with transforms
    if should_raise:
        with pytest.raises(NgioValueError):
            img2.require_rescalable(img1)
        return None
    zoom = ZoomTransform(
        input_image=img2,
        target_image=img1,
        order="nearest",
    )

    img1_data = img1.get_as_numpy()
    img2_data = img2.get_as_numpy(transforms=[zoom])
    assert img1_data.shape == img2_data.shape, (shape1, shape2, xy_pixelsize)
    img2.set_array(patch=img2_data, transforms=[zoom])

    roi = img1.build_image_roi_table().rois()[0]
    img2_roi_data = img2.get_roi_as_numpy(roi, transforms=[zoom])
    # Roi data should match exactly
    assert img1_data.shape == img2_roi_data.shape
    img2.set_roi(roi=roi, patch=img2_roi_data, transforms=[zoom])


def test_zoom_virtual_axes(
    tmp_path: Path,
):
    path1 = tmp_path / "image1.zarr"
    path2 = tmp_path / "image2.zarr"
    create_empty_ome_zarr(
        store=path1, shape=(3, 16, 16, 16), axes_names="czyx", xy_pixelsize=1.0
    )
    create_empty_ome_zarr(
        store=path2, shape=(16, 32, 32), axes_names="zyx", xy_pixelsize=0.5
    )
    img1 = open_image(path1)
    img2 = open_image(path2)

    # Also test with transforms
    zoom = ZoomTransform(
        input_image=img2,
        target_image=img1,
        order="nearest",
    )

    img1_data = img1.get_as_numpy()
    img2_data = img2.get_as_numpy(transforms=[zoom], axes_order="czyx")
    img2.set_array(patch=img2_data, transforms=[zoom], axes_order="czyx")
    assert img2_data.shape[0] == 1  # Virtual channel axis

    roi = img1.build_image_roi_table().rois()[0]
    img2_roi_data = img2.get_roi_as_numpy(roi, transforms=[zoom], axes_order="czyx")
    # Roi data should match exactly except for virtual axis
    assert img1_data.shape[1:] == img2_roi_data.shape[1:]
    img2.set_roi(roi=roi, patch=img2_roi_data, transforms=[zoom], axes_order="czyx")


def test_derive_image_consistency(
    tmp_path: Path,
):
    path1 = tmp_path / "image1.zarr"
    ome_zarr = create_empty_ome_zarr(
        store=path1,
        shape=(3, 16, 16, 16),
        axes_names="czyx",
        xy_pixelsize=1.0,
        levels=["s0", "s1", "s2"],
    )
    ome_zarr_der = ome_zarr.derive_image(tmp_path / "derived_image.zarr")
    assert ome_zarr.channel_labels == ome_zarr_der.channel_labels
    assert ome_zarr.levels_paths == ome_zarr_der.levels_paths
    assert ome_zarr.space_unit == ome_zarr_der.space_unit
    assert ome_zarr.time_unit == ome_zarr_der.time_unit

    label = ome_zarr.derive_label("derived_label")
    assert label.path == "s0"



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Generic, TypeVar

import numpy as np
import zarr
from dask.array import Array as DaskArray

from ngio.common._dimensions import Dimensions
from ngio.common._roi import Roi, RoiPixels
from ngio.io_pipes._ops_axes import (
    AxesOps,
    build_axes_ops,
    get_as_dask_axes_ops,
    get_as_numpy_axes_ops,
    set_as_dask_axes_ops,
    set_as_numpy_axes_ops,
)
from ngio.io_pipes._ops_slices import (
    SlicingInputType,
    SlicingOps,
    build_slicing_ops,
    get_slice_as_dask,
    get_slice_as_numpy,
    set_slice_as_dask,
    set_slice_as_numpy,
)
from ngio.io_pipes._ops_transforms import (
    TransformProtocol,
    get_as_dask_transform,
    get_as_numpy_transform,
    set_as_dask_transform,
    set_as_numpy_transform,
)


def setup_io_pipe(
    *,
    dimensions: Dimensions,
    slicing_dict: dict[str, SlicingInputType] | None = None,
    axes_order: Sequence[str] | None = None,
    remove_channel_selection: bool = False,
) -> tuple[SlicingOps, AxesOps]:
    """Setup the slicing tuple and axes ops for an IO pipe."""
    slicing_ops = build_slicing_ops(
        dimensions=dimensions,
        slicing_dict=slicing_dict,
        remove_channel_selection=remove_channel_selection,
    )

    axes_ops = build_axes_ops(
        dimensions=dimensions,
        input_axes=slicing_ops.slice_axes,
        axes_order=axes_order,
    )
    return slicing_ops, axes_ops


##############################################################
#
# "From Disk" Pipes
#
##############################################################

T = TypeVar("T")


class DataGetter(ABC, Generic[T]):
    def __init__(
        self,
        zarr_array: zarr.Array,
        slicing_ops: SlicingOps,
        axes_ops: AxesOps,
        transforms: Sequence[TransformProtocol] | None = None,
        roi: Roi | RoiPixels | None = None,
    ) -> None:
        self._zarr_array = zarr_array
        self._slicing_ops = slicing_ops
        self._axes_ops = axes_ops
        self._transforms = transforms
        self._roi = roi

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return (
            f"{name}(zarr_array={self._zarr_array}, "
            f"slicing_ops={self._slicing_ops}, "
            f"axes_ops={self._axes_ops}, "
            f"transforms={self._transforms})"
        )

    @property
    def zarr_array(self) -> zarr.Array:
        return self._zarr_array

    @property
    def slicing_ops(self) -> SlicingOps:
        return self._slicing_ops

    @property
    def axes_ops(self) -> AxesOps:
        return self._axes_ops

    @property
    def transforms(self) -> Sequence[TransformProtocol] | None:
        return self._transforms

    @property
    def roi(self) -> Roi | RoiPixels:
        if self._roi is None:
            name = self.__class__.__name__
            raise ValueError(f"No ROI defined for {name}.")
        return self._roi

    def __call__(self) -> T:
        return self.get()

    @abstractmethod
    def get(self) -> T:
        pass


class DataSetter(ABC, Generic[T]):
    def __init__(
        self,
        zarr_array: zarr.Array,
        slicing_ops: SlicingOps,
        axes_ops: AxesOps,
        transforms: Sequence[TransformProtocol] | None = None,
        roi: Roi | RoiPixels | None = None,
    ) -> None:
        self._zarr_array = zarr_array
        self._slicing_ops = slicing_ops
        self._axes_ops = axes_ops
        self._transforms = transforms
        self._roi = roi

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return (
            f"{name}(zarr_array={self._zarr_array}, "
            f"slicing_ops={self._slicing_ops}, "
            f"axes_ops={self._axes_ops}, "
            f"transforms={self._transforms})"
        )

    @property
    def zarr_array(self) -> zarr.Array:
        return self._zarr_array

    @property
    def slicing_ops(self) -> SlicingOps:
        return self._slicing_ops

    @property
    def axes_ops(self) -> AxesOps:
        return self._axes_ops

    @property
    def transforms(self) -> Sequence[TransformProtocol] | None:
        return self._transforms

    @property
    def roi(self) -> Roi | RoiPixels:
        if self._roi is None:
            name = self.__class__.__name__
            raise ValueError(f"No ROI defined for {name}.")
        return self._roi

    def __call__(self, patch: T) -> None:
        return self.set(patch)

    @abstractmethod
    def set(self, patch: T) -> None:
        pass


class NumpyGetter(DataGetter[np.ndarray]):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        remove_channel_selection: bool = False,
        roi: Roi | RoiPixels | None = None,
    ) -> None:
        """Build a pipe to get a numpy or dask array from a zarr array."""
        slicing_ops, axes_ops = setup_io_pipe(
            dimensions=dimensions,
            slicing_dict=slicing_dict,
            axes_order=axes_order,
            remove_channel_selection=remove_channel_selection,
        )
        super().__init__(
            zarr_array=zarr_array,
            slicing_ops=slicing_ops,
            axes_ops=axes_ops,
            transforms=transforms,
            roi=roi,
        )

    def get(self) -> np.ndarray:
        """Get a numpy array from the zarr array with ops."""
        array = get_slice_as_numpy(self._zarr_array, slicing_ops=self._slicing_ops)
        array = get_as_numpy_axes_ops(array, axes_ops=self._axes_ops)
        array = get_as_numpy_transform(
            array,
            slicing_ops=self._slicing_ops,
            axes_ops=self._axes_ops,
            transforms=self._transforms,
        )
        return array


class DaskGetter(DataGetter[DaskArray]):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        remove_channel_selection: bool = False,
        roi: Roi | RoiPixels | None = None,
    ) -> None:
        """Build a pipe to get a numpy or dask array from a zarr array."""
        slicing_ops, axes_ops = setup_io_pipe(
            dimensions=dimensions,
            slicing_dict=slicing_dict,
            axes_order=axes_order,
            remove_channel_selection=remove_channel_selection,
        )
        super().__init__(
            zarr_array=zarr_array,
            slicing_ops=slicing_ops,
            axes_ops=axes_ops,
            transforms=transforms,
            roi=roi,
        )

    def get(self) -> DaskArray:
        """Get a dask array from the zarr array with ops.

        The order of operations is:
        * get slice will load the data from the zarr array
        * get axes ops will reorder, squeeze or expand the axes
        * get transform will apply any additional transforms

        """
        array = get_slice_as_dask(self._zarr_array, slicing_ops=self._slicing_ops)
        array = get_as_dask_axes_ops(array, axes_ops=self._axes_ops)
        array = get_as_dask_transform(
            array,
            slicing_ops=self._slicing_ops,
            axes_ops=self._axes_ops,
            transforms=self._transforms,
        )
        return array


##############################################################
#
# "To Disk" Pipes
#
##############################################################


class NumpySetter(DataSetter[np.ndarray]):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        remove_channel_selection: bool = False,
        roi: Roi | RoiPixels | None = None,
    ) -> None:
        """Build a pipe to get a numpy or dask array from a zarr array."""
        slicing_ops, axes_ops = setup_io_pipe(
            dimensions=dimensions,
            slicing_dict=slicing_dict,
            axes_order=axes_order,
            remove_channel_selection=remove_channel_selection,
        )
        super().__init__(
            zarr_array=zarr_array,
            slicing_ops=slicing_ops,
            axes_ops=axes_ops,
            transforms=transforms,
            roi=roi,
        )

    def set(self, patch: np.ndarray) -> None:
        """Get a numpy array from the zarr array with ops."""
        patch = set_as_numpy_transform(
            array=patch,
            slicing_ops=self._slicing_ops,
            axes_ops=self._axes_ops,
            transforms=self._transforms,
        )
        patch = set_as_numpy_axes_ops(
            array=patch,
            axes_ops=self._axes_ops,
        )
        set_slice_as_numpy(
            zarr_array=self._zarr_array,
            patch=patch,
            slicing_ops=self._slicing_ops,
        )


class DaskSetter(DataSetter[DaskArray]):
    def __init__(
        self,
        *,
        zarr_array: zarr.Array,
        dimensions: Dimensions,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        slicing_dict: dict[str, SlicingInputType] | None = None,
        remove_channel_selection: bool = False,
        roi: Roi | RoiPixels | None = None,
    ) -> None:
        """Build a pipe to get a numpy or dask array from a zarr array."""
        slicing_ops, axes_ops = setup_io_pipe(
            dimensions=dimensions,
            slicing_dict=slicing_dict,
            axes_order=axes_order,
            remove_channel_selection=remove_channel_selection,
        )
        super().__init__(
            zarr_array=zarr_array,
            slicing_ops=slicing_ops,
            axes_ops=axes_ops,
            transforms=transforms,
            roi=roi,
        )

    def set(self, patch: DaskArray) -> None:
        """Get a dask array from the zarr array with ops."""
        patch = set_as_dask_transform(
            array=patch,
            slicing_ops=self._slicing_ops,
            axes_ops=self._axes_ops,
            transforms=self._transforms,
        )
        patch = set_as_dask_axes_ops(
            array=patch,
            axes_ops=self._axes_ops,
        )
        set_slice_as_dask(
            zarr_array=self._zarr_array,
            patch=patch,
            slicing_ops=self._slicing_ops,
        )



"""A module for handling label images in OME-NGFF files."""

from collections.abc import Sequence
from typing import Literal

from zarr.core.array import CompressorLike

from ngio.common import compute_masking_roi
from ngio.images._abstract_image import AbstractImage
from ngio.images._create import create_empty_label_container
from ngio.images._image import Image
from ngio.ome_zarr_meta import (
    LabelMetaHandler,
    NgioLabelMeta,
    PixelSize,
    find_label_meta_handler,
)
from ngio.ome_zarr_meta.ngio_specs import (
    DefaultSpaceUnit,
    DefaultTimeUnit,
    SpaceUnits,
    TimeUnits,
)
from ngio.tables import MaskingRoiTable
from ngio.utils import (
    NgioValidationError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)
from ngio.utils._zarr_utils import find_dimension_separator


class Label(AbstractImage[LabelMetaHandler]):
    """Placeholder class for a label."""

    get_as_numpy = AbstractImage._get_as_numpy
    get_as_dask = AbstractImage._get_as_dask
    get_array = AbstractImage._get_array
    get_roi_as_numpy = AbstractImage._get_roi_as_numpy
    get_roi_as_dask = AbstractImage._get_roi_as_dask
    get_roi = AbstractImage._get_roi
    set_array = AbstractImage._set_array
    set_roi = AbstractImage._set_roi

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: LabelMetaHandler | None,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.

        """
        if meta_handler is None:
            meta_handler = find_label_meta_handler(group_handler)
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )

    def __repr__(self) -> str:
        """Return the string representation of the label."""
        return f"Label(path={self.path}, {self.dimensions})"

    @property
    def meta(self) -> NgioLabelMeta:
        """Return the metadata."""
        return self._meta_handler.meta

    def set_axes_unit(
        self,
        space_unit: SpaceUnits = DefaultSpaceUnit,
        time_unit: TimeUnits = DefaultTimeUnit,
    ) -> None:
        """Set the axes unit of the image.

        Args:
            space_unit (SpaceUnits): The space unit of the image.
            time_unit (TimeUnits): The time unit of the image.
        """
        meta = self.meta
        meta = meta.to_units(space_unit=space_unit, time_unit=time_unit)
        self._meta_handler.write_meta(meta)

    def build_masking_roi_table(self) -> MaskingRoiTable:
        """Compute the masking ROI table."""
        return build_masking_roi_table(self)

    def consolidate(
        self,
        mode: Literal["dask", "numpy", "coarsen"] = "dask",
    ) -> None:
        """Consolidate the label on disk."""
        self._consolidate(
            order="nearest",
            mode=mode,
        )


class LabelsContainer:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler."""
        self._group_handler = group_handler
        # Validate the group
        # Either contains a labels attribute or is empty
        attrs = self._group_handler.load_attrs()
        if len(attrs) == 0:
            # It's an empty group
            pass
        elif "labels" in attrs and isinstance(attrs["labels"], list):
            # It's a valid group
            pass
        else:
            raise NgioValidationError(
                f"Invalid /labels group. "
                f"Expected a single labels attribute with a list of label names. "
                f"Found: {attrs}"
            )

    def list(self) -> list[str]:
        """Create the /labels group if it doesn't exist."""
        attrs = self._group_handler.load_attrs()
        return attrs.get("labels", [])

    def get(
        self,
        name: str,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        strict: bool = False,
    ) -> Label:
        """Get a label from the group.

        Args:
            name (str): The name of the label.
            path (str | None): The path to the image in the ome_zarr file.
            pixel_size (PixelSize | None): The pixel size of the image.
            strict (bool): Only used if the pixel size is provided. If True, the
                pixel size must match the image pixel size exactly. If False, the
                closest pixel size level will be returned.

        """
        if name not in self.list():
            raise NgioValueError(
                f"Label '{name}' not found in the Labels group. "
                f"Available labels: {self.list()}"
            )

        group_handler = self._group_handler.derive_handler(name)
        label_meta_handler = find_label_meta_handler(group_handler)
        path = label_meta_handler.meta.get_dataset(
            path=path, pixel_size=pixel_size, strict=strict
        ).path
        return Label(group_handler, path, label_meta_handler)

    def derive(
        self,
        name: str,
        ref_image: Image | Label,
        shape: Sequence[int] | None = None,
        pixel_size: PixelSize | None = None,
        axes_names: Sequence[str] | None = None,
        chunks: Sequence[int] | None = None,
        dtype: str = "uint32",
        dimension_separator: Literal[".", "/"] | None = None,
        compressors: CompressorLike | None = None,
        overwrite: bool = False,
    ) -> "Label":
        """Create an empty OME-Zarr label from a reference image.

        And add the label to the /labels group.

        Args:
            store (StoreOrGroup): The Zarr store or group to create the image in.
            ref_image (Image | Label): A reference image that will be used to create
                the new image.
            name (str): The name of the new image.
            shape (Sequence[int] | None): The shape of the new image.
            pixel_size (PixelSize | None): The pixel size of the new image.
            axes_names (Sequence[str] | None): The axes names of the new image.
                For labels, the channel axis is not allowed.
            chunks (Sequence[int] | None): The chunk shape of the new image.
            dtype (str): The data type of the new label.
            dimension_separator (DIMENSION_SEPARATOR | None): The separator to use for
                dimensions. If None it will use the same as the reference image.
            compressors (CompressorLike | None): The compressors to use. If None it will
                use the same as the reference image.
            overwrite (bool): Whether to overwrite an existing image.

        Returns:
            Label: The new label.

        """
        existing_labels = self.list()
        if name in existing_labels and not overwrite:
            raise NgioValueError(
                f"Label '{name}' already exists in the group. "
                "Use overwrite=True to replace it."
            )

        label_group = self._group_handler.get_group(name, create_mode=True)

        derive_label(
            store=label_group,
            ref_image=ref_image,
            name=name,
            shape=shape,
            pixel_size=pixel_size,
            axes_names=axes_names,
            chunks=chunks,
            dtype=dtype,
            dimension_separator=dimension_separator,
            compressors=compressors,
            overwrite=overwrite,
        )

        if name not in existing_labels:
            existing_labels.append(name)
            self._group_handler.write_attrs({"labels": existing_labels})

        return self.get(name)


def derive_label(
    store: StoreOrGroup,
    ref_image: Image | Label,
    name: str,
    shape: Sequence[int] | None = None,
    pixel_size: PixelSize | None = None,
    axes_names: Sequence[str] | None = None,
    chunks: Sequence[int] | None = None,
    dimension_separator: Literal[".", "/"] | None = None,
    compressors: CompressorLike | None = None,
    dtype: str = "uint32",
    overwrite: bool = False,
) -> None:
    """Create an empty OME-Zarr label from a reference image.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        ref_image (Image | Label): A reference image that will be used to
            create the new image.
        name (str): The name of the new image.
        shape (Sequence[int] | None): The shape of the new image.
        pixel_size (PixelSize | None): The pixel size of the new image.
        axes_names (Sequence[str] | None): The axes names of the new image.
            For labels, the channel axis is not allowed.
        chunks (Sequence[int] | None): The chunk shape of the new image.
        dtype (str): The data type of the new label.
        dimension_separator (DIMENSION_SEPARATOR | None): The separator to use for
            dimensions. If None it will use the same as the reference image.
        compressors (CompressorLike | None): The compressor to use. If None it will use
            the same as the reference image.
        overwrite (bool): Whether to overwrite an existing image.

    Returns:
        None

    """
    ref_meta = ref_image.meta

    if shape is None:
        shape = ref_image.shape

    if pixel_size is None:
        pixel_size = ref_image.pixel_size

    if axes_names is None:
        axes_names = ref_meta.axes_handler.axes_names
        c_axis = ref_meta.axes_handler.get_index("c")
    else:
        if "c" in axes_names:
            raise NgioValidationError(
                "Labels cannot have a channel axis. "
                "Please remove the channel axis from the axes names."
            )
        c_axis = None

    if len(axes_names) != len(shape):
        raise NgioValidationError(
            "The axes names of the new image does not match the reference image."
            f"Got {axes_names} for shape {shape}."
        )

    if chunks is None:
        chunks = ref_image.chunks

    if len(chunks) != len(shape):
        raise NgioValidationError(
            "The chunks of the new image does not match the reference image."
            f"Got {chunks} for shape {shape}."
        )

    if c_axis is not None:
        # remove channel if present
        shape = list(shape)
        shape = shape[:c_axis] + shape[c_axis + 1 :]
        chunks = list(chunks)
        chunks = chunks[:c_axis] + chunks[c_axis + 1 :]
        axes_names = list(axes_names)
        axes_names = axes_names[:c_axis] + axes_names[c_axis + 1 :]

    if dimension_separator is None:
        dimension_separator = find_dimension_separator(ref_image.zarr_array)
    if compressors is None:
        compressors = ref_image.zarr_array.compressors  # type: ignore

    _ = create_empty_label_container(
        store=store,
        shape=shape,
        pixelsize=ref_image.pixel_size.x,
        z_spacing=ref_image.pixel_size.z,
        time_spacing=ref_image.pixel_size.t,
        levels=ref_meta.paths,
        yx_scaling_factor=ref_meta.yx_scaling(),
        z_scaling_factor=ref_meta.z_scaling(),
        time_unit=ref_image.pixel_size.time_unit,
        space_unit=ref_image.pixel_size.space_unit,
        axes_names=axes_names,
        chunks=chunks,
        dtype=dtype,
        dimension_separator=dimension_separator,
        compressors=compressors,
        overwrite=overwrite,
        version=ref_meta.version,
        name=name,
    )
    return None


def build_masking_roi_table(label: Label) -> MaskingRoiTable:
    """Compute the masking ROI table for a label."""
    array = label.get_as_dask(axes_order=["t", "z", "y", "x"])
    rois = compute_masking_roi(array, label.pixel_size)
    return MaskingRoiTable(rois, reference_label=label.meta.name)



"""Module to handle the channel information in the metadata.

Stores the same information as the Omero section of the ngff 0.4 metadata.
"""

from collections.abc import Sequence
from difflib import SequenceMatcher
from enum import Enum
from typing import Any, TypeVar

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from ngio.utils import NgioValidationError, NgioValueError

################################################################################################
#
# Omero Section of the Metadata is used to store channel information and visualisation
# settings.
# This section is transitory and will be likely changed in the future.
#
#################################################################################################


class NgioColors(str, Enum):
    """Default colors for the channels."""

    dapi = "0000FF"
    hoechst = "0000FF"
    gfp = "00FF00"
    cy3 = "FFFF00"
    cy5 = "FF0000"
    brightfield = "808080"
    red = "FF0000"
    yellow = "FFFF00"
    magenta = "FF00FF"
    cyan = "00FFFF"
    gray = "808080"
    green = "00FF00"

    @staticmethod
    def semi_random_pick(channel_name: str | None = None) -> "NgioColors":
        """Try to fuzzy match the color to the channel name.

        - If a channel name is given will try to match the channel name to the color.
        - If name has the paatern 'channel_x' cyclic rotate over a list of colors
            [cyan, magenta, yellow, green]
        - If no channel name is given will return a random color.
        """
        available_colors = NgioColors._member_names_

        if channel_name is None:
            # Purely random color
            color_str = available_colors[np.random.randint(0, len(available_colors))]
            return NgioColors.__members__[color_str]

        if channel_name.startswith("channel_"):
            # Rotate over a list of colors
            defaults_colors = [
                NgioColors.cyan,
                NgioColors.magenta,
                NgioColors.yellow,
                NgioColors.green,
            ]

            try:
                index = int(channel_name.split("_")[-1]) % len(defaults_colors)
                return defaults_colors[index]
            except ValueError:
                # If the name of the channel is something like
                # channel_dapi this will fail an proceed to the
                # standard fuzzy match
                pass

        similarity = {}
        for color in available_colors:
            # try to match the color to the channel name
            similarity[color] = SequenceMatcher(None, channel_name, color).ratio()
        # Get the color with the highest similarity
        color_str = max(similarity, key=similarity.get)  # type: ignore (max type overload fails to infer type)
        assert isinstance(color_str, str), "Color name must be a string."
        return NgioColors.__members__[color_str]


def valid_hex_color(v: str) -> bool:
    """Validate a hexadecimal color.

    Check that `color` is made of exactly six elements which are letters
    (a-f or A-F) or digits (0-9).

    Implementation source:
    https://github.com/fractal-analytics-platform/fractal-tasks-core/fractal_tasks_core/channels.py#L87
    Original authors:
     - Tommaso Comparin <tommaso.comparin@exact-lab.it>
    """
    if len(v) != 6:
        return False
    allowed_characters = "abcdefABCDEF0123456789"
    for character in v:
        if character not in allowed_characters:
            return False
    return True


def into_valid_hex_color(v: str) -> str:
    """Convert a string into a valid hexadecimal color.

    If the string is already a valid hexadecimal color, return it.
    Otherwise, return a hexadecimal color based on the hash of the string.
    """
    # strip leading '#' if present
    v = v.lstrip("#")
    if valid_hex_color(v):
        return v

    return NgioColors.semi_random_pick(v.lower()).value


class ChannelVisualisation(BaseModel):
    """Channel visualisation model.

    Contains the information about the visualisation of a channel.

    Attributes:
        color(str): The color of the channel in hexadecimal format or a color name.
        min(int | float): The minimum value of the channel.
        max(int | float): The maximum value of the channel.
        start(int | float): The start value of the channel.
        end(int | float): The end value of the channel.
        active(bool): Whether the channel is active.
    """

    color: str | NgioColors | None = Field(default=None, validate_default=True)
    min: int | float = 0
    max: int | float = 65535
    start: int | float = 0
    end: int | float = 65535
    active: bool = True
    model_config = ConfigDict(extra="allow", frozen=True)

    @field_validator("color", mode="after")
    def validate_color(cls, value: str | NgioColors) -> str:
        """Color validator.

        There are three possible values to set a color:
         - A hexadecimal string.
         - A color name.
         - A NgioColors element.
        """
        if value is None:
            return NgioColors.semi_random_pick().value
        if isinstance(value, str):
            return into_valid_hex_color(value)
        elif isinstance(value, NgioColors):
            return value.value
        else:
            raise NgioValueError(f"Invalid color {value}.")

    @model_validator(mode="before")
    def check_start_end(cls, data):
        """Check that the start and end values are valid.

        If the start and end values are equal, set the end value to start + 1
        """
        start = data.get("start", None)
        end = data.get("end", None)
        if start is None or end is None:
            return data
        if abs(end - start) < 1e-6:
            data["end"] = start + 1
        return data

    @classmethod
    def default_init(
        cls,
        color: str | NgioColors | None = None,
        start: int | float | None = None,
        end: int | float | None = None,
        active: bool = True,
        data_type: Any = np.uint16,
    ) -> "ChannelVisualisation":
        """Create a ChannelVisualisation object with the default unit.

        Args:
            color(str): The color of the channel in hexadecimal format or a color name.
            start(int | float | None): The start value of the channel.
            end(int | float | None): The end value of the channel.
            data_type(Any): The data type of the channel.
            active(bool): Whether the channel should be shown by default.
        """
        for func in [np.iinfo, np.finfo]:
            try:
                min_value = func(data_type).min
                max_value = func(data_type).max
                break
            except ValueError:
                continue
        else:
            raise NgioValueError(f"Invalid data type {data_type}.")

        start = start if start is not None else min_value
        end = end if end is not None else max_value
        return cls(
            color=color,
            min=min_value,
            max=max_value,
            start=start,
            end=end,
            active=active,
        )

    @property
    def valid_color(self) -> str:
        """Return the valid color."""
        if isinstance(self.color, NgioColors):
            return self.color.value
        elif isinstance(self.color, str):
            return self.color
        else:
            raise NgioValueError(f"Invalid color {self.color}.")


def default_channel_name(index: int) -> str:
    """Return the default channel name."""
    return f"channel_{index}"


class Channel(BaseModel):
    """Information about a channel in the image.

    Attributes:
        label(str): The label of the channel.
        wavelength_id(str): The wavelength ID of the channel.
        extra_fields(dict): To reduce the api surface, extra fields are stored in the
            the channel attributes will be stored in the extra_fields attribute.
    """

    label: str
    wavelength_id: str | None = None
    channel_visualisation: ChannelVisualisation
    model_config = ConfigDict(extra="allow", frozen=True)

    @classmethod
    def default_init(
        cls,
        label: str,
        wavelength_id: str | None = None,
        color: str | NgioColors | None = None,
        start: int | float | None = None,
        end: int | float | None = None,
        active: bool = True,
        data_type: Any = np.uint16,
    ) -> "Channel":
        """Create a Channel object with the default unit.

        Args:
            label(str): The label of the channel.
            wavelength_id(str | None): The wavelength ID of the channel.
            color(str): The color of the channel in hexadecimal format or a color name.
                If None, the color will be picked based on the label.
            start(int | float | None): The start value of the channel.
            end(int | float | None): The end value of the channel.
            active(bool): Whether the channel should be shown by default.
            data_type(Any): The data type of the channel.
        """
        if color is None:
            # If no color is provided, try to pick a color based on the label
            # See the NgioColors.semi_random_pick method for more details.
            color = label

        channel_visualization = ChannelVisualisation.default_init(
            color=color, start=start, end=end, active=active, data_type=data_type
        )

        if wavelength_id is None:
            # TODO Evaluate if a better default value can be used
            wavelength_id = label

        return cls(
            label=label,
            wavelength_id=wavelength_id,
            channel_visualisation=channel_visualization,
        )


T = TypeVar("T")


def _check_elements(elements: Sequence[T], expected_type: Any) -> Sequence[T]:
    """Check that the elements are of the same type."""
    if len(elements) == 0:
        raise NgioValidationError("At least one element must be provided.")

    for element in elements:
        if not isinstance(element, expected_type):
            raise NgioValidationError(
                f"All elements must be of the same type {expected_type}. Got {element}."
            )

    return elements


def _check_unique(elements: Sequence[T]) -> Sequence[T]:
    """Check that the elements are unique."""
    if len(set(elements)) != len(elements):
        raise NgioValidationError("All elements must be unique.")
    return elements


class ChannelsMeta(BaseModel):
    """Information about the channels in the image.

    This model is roughly equivalent to the Omero section of the ngff 0.4 metadata.

    Attributes:
        channels(list[Channel]): The list of channels in the image.
    """

    channels: list[Channel] = Field(default_factory=list)
    model_config = ConfigDict(extra="allow", frozen=True)

    @field_validator("channels", mode="after")
    def validate_channels(cls, value: list[Channel]) -> list[Channel]:
        """Check that the channels are unique."""
        _check_unique([ch.label for ch in value])
        return value

    @classmethod
    def default_init(
        cls,
        labels: Sequence[str | None] | int,
        wavelength_id: Sequence[str | None] | None = None,
        colors: Sequence[str | NgioColors | None] | None = None,
        start: Sequence[int | float | None] | int | float | None = None,
        end: Sequence[int | float | None] | int | float | None = None,
        active: Sequence[bool | None] | None = None,
        data_type: Any = np.uint16,
        **omero_kwargs: dict,
    ) -> "ChannelsMeta":
        """Create a ChannelsMeta object with the default unit.

        Args:
            labels(Sequence[str | None] | int): The list of channels names
                in the image. If an integer is provided, the channels will be
                named "channel_i".
            wavelength_id(Sequence[str | None] | None): The wavelength ID of the
                channel. If None, the wavelength ID will be the same as the
                channel name.
            colors(Sequence[str | NgioColors | None] | None): The list of
                colors for the channels. If None, the colors will be random.
            start(Sequence[int | float | None] | int | float | None): The start
                value of the channel. If None, the start value will be the
                minimum value of the data type.
            end(Sequence[int | float | None] | int | float | None): The end
                value of the channel. If None, the end value will be the
                maximum value of the data type.
            data_type(Any): The data type of the channel. Will be used to set the
                min and max values of the channel.
            active (Sequence[bool | None] | None): Whether the channel should
                be shown by default.
            omero_kwargs(dict): Extra fields to store in the omero attributes.
        """
        if isinstance(labels, int):
            labels = [default_channel_name(i) for i in range(labels)]

        labels = _check_elements(labels, str)
        labels = _check_unique(labels)

        _wavelength_id: Sequence[str | None] = [None] * len(labels)
        if wavelength_id is None:
            _wavelength_id: Sequence[str | None] = [None] * len(labels)
        else:
            _wavelength_id = _check_elements(wavelength_id, str)
            _wavelength_id = _check_unique(wavelength_id)

        if colors is None:
            _colors = [NgioColors.semi_random_pick(label) for label in labels]
        else:
            _colors = _check_elements(colors, str | NgioColors)

        if start is None:
            _start = [None] * len(labels)
        elif isinstance(start, int | float):
            _start = [start] * len(labels)
        else:
            _start = _check_elements(start, (int, float))

        if end is None:
            _end = [None] * len(labels)
        elif isinstance(end, int | float):
            _end = [end] * len(labels)
        else:
            _end = _check_elements(end, (int, float))

        if active is None:
            _active = [True] * len(labels)
        else:
            _active = _check_elements(active, (bool,))

        all_lengths = [
            len(labels),
            len(_wavelength_id),
            len(_colors),
            len(_start),
            len(_end),
            len(_active),
        ]
        if len(set(all_lengths)) != 1:
            raise NgioValueError("Channels information must all have the same length.")

        channels = []
        for ch_name, w_id, color, s, e, a in zip(
            labels, _wavelength_id, _colors, _start, _end, _active, strict=True
        ):
            channels.append(
                Channel.default_init(
                    label=ch_name,
                    wavelength_id=w_id,
                    color=color,
                    start=s,
                    end=e,
                    active=a,
                    data_type=data_type,
                )
            )
        return cls(channels=channels, **omero_kwargs)

    @property
    def channel_labels(self) -> list[str]:
        """Get the labels of the channels in the image."""
        return [channel.label for channel in self.channels]

    @property
    def channel_wavelength_ids(self) -> list[str | None]:
        """Get the wavelength IDs of the channels in the image."""
        return [channel.wavelength_id for channel in self.channels]

    def get_channel_idx(
        self, channel_label: str | None = None, wavelength_id: str | None = None
    ) -> int:
        """Get the index of a channel by its label or wavelength ID."""
        # Only one of the arguments must be provided
        if channel_label is not None and wavelength_id is not None:
            raise NgioValueError(
                "get_channel_idx must receive either label or wavelength_id, not both."
            )

        if channel_label is not None:
            if channel_label not in self.channel_labels:
                raise NgioValueError(f"Channel with label {channel_label} not found.")
            return self.channel_labels.index(channel_label)

        if wavelength_id is not None:
            if wavelength_id not in self.channel_wavelength_ids:
                raise NgioValueError(
                    f"Channel with wavelength ID {wavelength_id} not found."
                )
            return self.channel_wavelength_ids.index(wavelength_id)

        raise NgioValueError(
            "get_channel_idx must receive either label or wavelength_id"
        )



"""Models for programmatic description of image resources."""

from pathlib import Path

from pydantic import BaseModel, Field

from ngio.ome_zarr_meta.ngio_specs import (
    DefaultSpaceUnit,
    DefaultTimeUnit,
    SpaceUnits,
    TimeUnits,
)


class LabelsInfo(BaseModel):
    """Metadata for a label image."""

    name: str
    label_path: Path
    ensure_unique_labels: bool = True
    create_masking_table: bool = False
    dtype: str = "uint32"


class SampleInfo(BaseModel):
    """Metadata necessary to create an OME-Ngff from image files."""

    img_path: Path
    labels: list[LabelsInfo] = Field(default_factory=list)
    xy_pixelsize: float
    z_spacing: float = 1.0
    time_spacing: float = 1.0
    space_unit: SpaceUnits = DefaultSpaceUnit
    time_unit: TimeUnits = DefaultTimeUnit
    name: str | None = None
    info: str | None = None



"""Implementation of the FeatureTableV1 class.

This class follows the roi_table specification at:
https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
"""

from typing import Literal

from pydantic import BaseModel, Field

from ngio.tables._abstract_table import AbstractBaseTable
from ngio.tables.backends import BackendMeta, TableBackend, TabularData
from ngio.utils import NgioValueError
from ngio.utils._zarr_utils import ZarrGroupHandler


class RegionMeta(BaseModel):
    """Metadata for the region."""

    path: str


class FeatureTableMeta(BackendMeta):
    """Metadata for the ROI table."""

    table_version: Literal["1"] = "1"
    type: Literal["feature_table"] = "feature_table"
    region: RegionMeta | None = None
    instance_key: str = "label"  # Legacy field, kept for compatibility
    # Backend metadata
    index_key: str | None = "label"
    index_type: Literal["int", "str"] | None = "int"
    # Columns optional types
    categorical_columns: list[str] = Field(default_factory=list)
    measurement_columns: list[str] = Field(default_factory=list)
    metadata_columns: list[str] = Field(default_factory=list)


class FeatureTableV1(AbstractBaseTable):
    def __init__(
        self,
        table_data: TabularData | None = None,
        *,
        reference_label: str | None = None,
        meta: FeatureTableMeta | None = None,
    ) -> None:
        """Initialize the GenericTable."""
        if meta is None:
            meta = FeatureTableMeta()

        if reference_label is not None:
            path = f"../labels/{reference_label}"
            meta = FeatureTableMeta(region=RegionMeta(path=path))

        if table_data is not None and not isinstance(table_data, TabularData):
            raise NgioValueError(
                f"The table is not of type SupportedTables. Got {type(table_data)}"
            )

        if meta.index_key is None:
            meta.index_key = "label"

        if meta.index_type is None:
            meta.index_type = "int"

        meta.instance_key = meta.index_key

        super().__init__(
            table_data=table_data,
            meta=meta,
        )

    def __repr__(self) -> str:
        """Return a string representation of the table."""
        num_rows = len(self.dataframe) if self.dataframe is not None else 0
        num_columns = len(self.dataframe.columns) if self.dataframe is not None else 0
        properties = f"num_rows={num_rows}, num_columns={num_columns}"
        if self.reference_label is not None:
            properties += f", reference_label={self.reference_label}"
        return f"FeatureTableV1({properties})"

    @classmethod
    def from_handler(
        cls,
        handler: ZarrGroupHandler,
        backend: TableBackend | None = None,
    ) -> "FeatureTableV1":
        return cls._from_handler(
            handler=handler,
            backend=backend,
            meta_model=FeatureTableMeta,
        )

    @property
    def meta(self) -> FeatureTableMeta:
        """Return the metadata of the table."""
        if not isinstance(self._meta, FeatureTableMeta):
            raise NgioValueError(
                "The metadata of the table is not of type FeatureTableMeta."
            )
        return self._meta

    @staticmethod
    def table_type() -> str:
        """Return the type of the table."""
        return "feature_table"

    @staticmethod
    def version() -> str:
        """The generic table does not have a version.

        Since does not follow a specific schema.
        """
        return "1"

    @property
    def reference_label(self) -> str | None:
        """Return the reference label."""
        path = self.meta.region
        if path is None:
            return None

        path = path.path
        path = path.split("/")[-1]
        return path



import asyncio
from pathlib import Path
from typing import Literal

import pandas as pd
import pytest

from ngio import OmeZarrContainer, create_empty_ome_zarr
from ngio.images import (
    concatenate_image_tables,
    concatenate_image_tables_as,
    concatenate_image_tables_as_async,
    concatenate_image_tables_async,
    list_image_tables,
    list_image_tables_async,
)
from ngio.tables import FeatureTable, GenericTable


def create_sample_ome_zarr(
    tmp_path: Path, name: str, tables: list[str]
) -> OmeZarrContainer:
    store = tmp_path / f"{name}.zarr"
    ome_zarr_container = create_empty_ome_zarr(
        store=store,
        shape=(32, 32),
        xy_pixelsize=0.1,
    )
    for table_name in tables:
        table_data = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6], "label": [1, 2, 3]})
        table = FeatureTable(table_data=table_data)
        ome_zarr_container.add_table(table_name, table, backend="json")
    return ome_zarr_container


def test_list_sync_api(tmp_path: Path):
    ome_zarr_1 = create_sample_ome_zarr(tmp_path, "test1", ["table1", "table2"])
    ome_zarr_2 = create_sample_ome_zarr(tmp_path, "test2", ["table1"])

    assert list_image_tables([ome_zarr_1, ome_zarr_2], mode="common") == ["table1"]
    assert list_image_tables([ome_zarr_1, ome_zarr_2], mode="all") == [
        "table1",
        "table2",
    ]


def test_list_async_api(tmp_path: Path):
    ome_zarr_1 = create_sample_ome_zarr(tmp_path, "test1", ["table1", "table2"])
    ome_zarr_2 = create_sample_ome_zarr(tmp_path, "test2", ["table1"])

    assert asyncio.run(
        list_image_tables_async([ome_zarr_1, ome_zarr_2], mode="common")
    ) == ["table1"]
    assert asyncio.run(
        list_image_tables_async([ome_zarr_1, ome_zarr_2], mode="all")
    ) == [
        "table1",
        "table2",
    ]


@pytest.mark.parametrize(
    "table, mode, strict",
    [
        ("table1", "eager", True),
        ("table1", "lazy", True),
        ("table2", "eager", False),
        ("table2", "eager", True),
    ],
)
def test_cat_sync_api(
    tmp_path: Path, table: str, mode: Literal["eager", "lazy"], strict: bool
):
    ome_zarr_1 = create_sample_ome_zarr(tmp_path, "test1", ["table1", "table2"])
    ome_zarr_2 = create_sample_ome_zarr(tmp_path, "test2", ["table1"])

    extras1 = {"column1": "value1"}
    extras2 = {"column1": "value2"}
    if strict and table == "table2":
        with pytest.raises(ValueError):
            concatenate_image_tables(
                [ome_zarr_1, ome_zarr_2],
                extras=[extras1, extras2],
                name=table,
                mode=mode,
                strict=strict,
            )
        return None

    concatenated_table = concatenate_image_tables(
        [ome_zarr_1, ome_zarr_2],
        extras=[extras1, extras2],
        name=table,
        mode=mode,
        strict=strict,
    )
    assert isinstance(concatenated_table, FeatureTable)

    df = concatenated_table.dataframe
    df = df.reset_index()
    assert set(df.columns) == {"x", "y", "label", "column1"}
    if "table2" in table:
        assert df.shape == (3, 4), df.shape
    else:
        assert df.shape == (6, 4), df.shape


def test_cat_as_sync(tmp_path: Path):
    ome_zarr_1 = create_sample_ome_zarr(tmp_path, "test1", ["table1", "table2"])
    ome_zarr_2 = create_sample_ome_zarr(tmp_path, "test2", ["table1"])

    extras1 = {"column1": "value1"}
    extras2 = {"column1": "value2"}

    concatenated_table = concatenate_image_tables_as(
        [ome_zarr_1, ome_zarr_2],
        extras=[extras1, extras2],
        name="table1",
        table_cls=GenericTable,
    )

    assert isinstance(concatenated_table, GenericTable)


def test_set_index(tmp_path: Path):
    ome_zarr_1 = create_sample_ome_zarr(tmp_path, "test1", ["table1", "table2"])
    ome_zarr_2 = create_sample_ome_zarr(tmp_path, "test2", ["table1"])

    extras1 = {"column1": "value1"}
    extras2 = {"column1": "value2"}

    concatenated_table = concatenate_image_tables_as(
        [ome_zarr_1, ome_zarr_2],
        extras=[extras1, extras2],
        name="table1",
        table_cls=GenericTable,
        index_key="Index",
    )
    df = concatenated_table.dataframe
    assert set(df.columns) == {"x", "y", "label", "column1"}
    assert df.index.name == "Index"


def test_cat_async_api(tmp_path: Path):
    ome_zarr_1 = create_sample_ome_zarr(tmp_path, "test1", ["table1", "table2"])
    ome_zarr_2 = create_sample_ome_zarr(tmp_path, "test2", ["table1"])

    extras1 = {"column1": "value1"}
    extras2 = {"column1": "value2"}

    concatenated_table = asyncio.run(
        concatenate_image_tables_async(
            [ome_zarr_1, ome_zarr_2],
            extras=[extras1, extras2],
            name="table1",
        )
    )
    assert isinstance(concatenated_table, FeatureTable)

    df = concatenated_table.dataframe
    df = df.reset_index()
    assert set(df.columns) == {"x", "y", "label", "column1"}
    assert df.shape == (6, 4), df.shape

    concatenate_table = asyncio.run(
        concatenate_image_tables_as_async(
            [ome_zarr_1, ome_zarr_2],
            extras=[extras1, extras2],
            name="table1",
            table_cls=GenericTable,
        )
    )
    assert isinstance(concatenate_table, GenericTable)



import pytest
import zarr

from ngio.common import Dimensions
from ngio.io_pipes._ops_slices import (
    build_slicing_ops,
    get_slice_as_dask,
    get_slice_as_numpy,
    set_slice_as_dask,
    set_slice_as_numpy,
)
from ngio.io_pipes._ops_slices_utils import check_if_regions_overlap
from ngio.ome_zarr_meta import AxesHandler, Dataset
from ngio.ome_zarr_meta.ngio_specs import Axis


@pytest.mark.parametrize(
    "input_axes,output_axes,input_shape,output_shape,slicing_dict",
    [
        ("tczyx", "zyx", (1, 1, 10, 11, 12), (10, 11, 12), {"t": 0, "c": 0}),
        ("czyx", "czyx", (3, 10, 11, 12), (2, 10, 11, 12), {"c": (0, 2)}),
        ("czyx", "czy", (3, 10, 11, 12), (2, 10, 11), {"c": (0, 2), "x": 4}),
        ("zyx", "zyx", (10, 11, 12), (10, 11, 12), {"t": slice(None), "c": 0}),
    ],
)
def test_slicing_ops_base(
    input_axes: str,
    output_axes: str,
    input_shape: tuple[int, ...],
    output_shape: tuple[int, ...],
    slicing_dict: dict[str, int | slice | tuple[int, ...]],
):
    axes = [Axis(name=name) for name in input_axes]

    ax_mapper = AxesHandler(axes=axes)
    ds = Dataset(
        path="0",
        axes_handler=ax_mapper,
        scale=[1.0] * len(axes),
        translation=[0.0] * len(axes),
    )
    dims = Dimensions(shape=input_shape, chunks=input_shape, dataset=ds)

    slicing_ops = build_slicing_ops(
        dimensions=dims,
        slicing_dict=slicing_dict,  # type: ignore[arg-type]
        remove_channel_selection=False,
    )
    assert slicing_ops.slice_axes == tuple(output_axes)
    assert slicing_ops.slice_chunks() == {(0,) * len(slicing_ops.on_disk_axes)}

    # Numpy
    zarr_data = zarr.zeros(shape=input_shape, chunks=input_shape, dtype="uint8")
    data = get_slice_as_numpy(zarr_data, slicing_ops=slicing_ops)
    assert data.shape == output_shape
    set_slice_as_numpy(zarr_array=zarr_data, patch=data, slicing_ops=slicing_ops)

    # Dask
    zarr_data = zarr.zeros(shape=input_shape, chunks=input_shape, dtype="uint8")
    data = get_slice_as_dask(zarr_data, slicing_ops=slicing_ops)
    assert data.shape == output_shape
    set_slice_as_dask(zarr_array=zarr_data, patch=data, slicing_ops=slicing_ops)


def test_chunk_slice():
    input_axes = "tczyx"
    input_shape = (2, 3, 10, 11, 12)
    chunk_shape = (1, 1, 10, 11, 12)
    axes = [Axis(name=name) for name in input_axes]

    ax_mapper = AxesHandler(axes=axes)
    ds = Dataset(
        path="0",
        axes_handler=ax_mapper,
        scale=[1.0] * len(axes),
        translation=[0.0] * len(axes),
    )
    dims = Dimensions(shape=input_shape, chunks=chunk_shape, dataset=ds)

    slicing_dict = {"t": 0, "c": (0, 2)}
    slicing_ops1 = build_slicing_ops(
        dimensions=dims,
        slicing_dict=slicing_dict,  # type: ignore[arg-type]
        remove_channel_selection=False,
    )
    assert slicing_ops1.slice_chunks() == {(0, 0, 0, 0, 0), (0, 2, 0, 0, 0)}

    slicing_dict = {"t": 1, "c": (0, 2)}
    slicing_ops2 = build_slicing_ops(
        dimensions=dims,
        slicing_dict=slicing_dict,  # type: ignore[arg-type]
        remove_channel_selection=False,
    )
    assert slicing_ops2.slice_chunks() == {(1, 0, 0, 0, 0), (1, 2, 0, 0, 0)}

    # Check for overlapping regions
    assert not check_if_regions_overlap(
        [slicing_ops1.normalized_slicing_tuple, slicing_ops2.normalized_slicing_tuple]
    )

    # Repeat with bigger tuples as different element
    slicing_dict = {"t": 0, "c": (0, 2)}
    slicing_ops2 = build_slicing_ops(
        dimensions=dims,
        slicing_dict=slicing_dict,  # type: ignore[arg-type]
        remove_channel_selection=False,
    )

    # Check for overlapping regions
    assert check_if_regions_overlap(
        [slicing_ops1.normalized_slicing_tuple, slicing_ops2.normalized_slicing_tuple]
    )



"""Tables implementations for fractal_tables v1."""

from ngio.tables.v1._condition_table import ConditionTableMeta, ConditionTableV1
from ngio.tables.v1._feature_table import FeatureTableMeta, FeatureTableV1
from ngio.tables.v1._generic_table import GenericTable
from ngio.tables.v1._roi_table import (
    MaskingRoiTableV1,
    MaskingRoiTableV1Meta,
    RoiTableV1,
    RoiTableV1Meta,
)

__all__ = [
    "ConditionTableMeta",
    "ConditionTableV1",
    "FeatureTableMeta",
    "FeatureTableV1",
    "GenericTable",
    "MaskingRoiTableV1",
    "MaskingRoiTableV1Meta",
    "RoiTableV1",
    "RoiTableV1Meta",
]



from abc import ABC, abstractmethod
from collections.abc import Callable, Generator
from typing import Generic, Literal, Self, TypeVar, overload

from ngio import Roi
from ngio.experimental.iterators._mappers import BasicMapper, MapperProtocol
from ngio.experimental.iterators._rois_utils import (
    by_chunks,
    by_yx,
    by_zyx,
    grid,
    rois_product,
)
from ngio.images._abstract_image import AbstractImage
from ngio.io_pipes._io_pipes_types import DataGetterProtocol, DataSetterProtocol
from ngio.io_pipes._ops_slices_utils import check_if_regions_overlap
from ngio.tables import GenericRoiTable
from ngio.utils import NgioValueError

NumpyPipeType = TypeVar("NumpyPipeType")
DaskPipeType = TypeVar("DaskPipeType")


class AbstractIteratorBuilder(ABC, Generic[NumpyPipeType, DaskPipeType]):
    """Base class for building iterators over ROIs."""

    _rois: list[Roi]
    _ref_image: AbstractImage

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(regions={len(self._rois)})"

    @abstractmethod
    def get_init_kwargs(self) -> dict:
        """Return the initialization arguments for the iterator.

        This is used to clone the iterator with the same parameters
        after every "product" operation.
        """
        pass

    @property
    def rois(self) -> list[Roi]:
        """Get the list of ROIs for the iterator."""
        return self._rois

    def _set_rois(self, rois: list[Roi]) -> None:
        """Set the list of ROIs for the iterator."""
        self._rois = rois

    @property
    def ref_image(self) -> AbstractImage:
        """Get the reference image for the iterator."""
        return self._ref_image

    def _new_from_rois(self, rois: list[Roi]) -> Self:
        """Create a new instance of the iterator with a different set of ROIs."""
        init_kwargs = self.get_init_kwargs()
        new_instance = self.__class__(**init_kwargs)
        new_instance._set_rois(rois)
        return new_instance

    def grid(
        self,
        size_x: int | None = None,
        size_y: int | None = None,
        size_z: int | None = None,
        size_t: int | None = None,
        stride_x: int | None = None,
        stride_y: int | None = None,
        stride_z: int | None = None,
        stride_t: int | None = None,
        base_name: str = "",
    ) -> Self:
        """Create a grid of ROIs based on the input image dimensions."""
        rois = grid(
            rois=self.rois,
            ref_image=self.ref_image,
            size_x=size_x,
            size_y=size_y,
            size_z=size_z,
            size_t=size_t,
            stride_x=stride_x,
            stride_y=stride_y,
            stride_z=stride_z,
            stride_t=stride_t,
            base_name=base_name,
        )
        return self._new_from_rois(rois)

    def by_yx(self) -> Self:
        """Return a new iterator that iterates over ROIs by YX coordinates."""
        rois = by_yx(self.rois, self.ref_image)
        return self._new_from_rois(rois)

    def by_zyx(self, strict: bool = True) -> Self:
        """Return a new iterator that iterates over ROIs by ZYX coordinates.

        Args:
            strict (bool): If True, only iterate over ZYX if a Z axis
                is present and not of size 1.

        """
        rois = by_zyx(self.rois, self.ref_image, strict=strict)
        return self._new_from_rois(rois)

    def by_chunks(self, overlap_xy: int = 0, overlap_z: int = 0) -> Self:
        """Return a new iterator that iterates over ROIs by chunks.

        Args:
            overlap_xy (int): Overlap in XY dimensions.
            overlap_z (int): Overlap in Z dimension.

        Returns:
            SegmentationIterator: A new iterator with chunked ROIs.
        """
        rois = by_chunks(
            self.rois, self.ref_image, overlap_xy=overlap_xy, overlap_z=overlap_z
        )
        return self._new_from_rois(rois)

    def product(self, other: list[Roi] | GenericRoiTable) -> Self:
        """Cartesian product of the current ROIs with an arbitrary list of ROIs."""
        if isinstance(other, GenericRoiTable):
            other = other.rois()
        rois = rois_product(self.rois, other)
        return self._new_from_rois(rois)

    @abstractmethod
    def build_numpy_getter(self, roi: Roi) -> DataGetterProtocol[NumpyPipeType]:
        """Build a getter function for the given ROI."""
        raise NotImplementedError

    @abstractmethod
    def build_numpy_setter(self, roi: Roi) -> DataSetterProtocol[NumpyPipeType] | None:
        """Build a setter function for the given ROI."""
        raise NotImplementedError

    @abstractmethod
    def build_dask_getter(self, roi: Roi) -> DataGetterProtocol[DaskPipeType]:
        """Build a Dask reader function for the given ROI."""
        raise NotImplementedError

    @abstractmethod
    def build_dask_setter(self, roi: Roi) -> DataSetterProtocol[DaskPipeType] | None:
        """Build a Dask setter function for the given ROI."""
        raise NotImplementedError

    @abstractmethod
    def post_consolidate(self) -> None:
        """Post-process the consolidated data."""
        raise NotImplementedError

    def _numpy_getters_generator(self) -> Generator[DataGetterProtocol[NumpyPipeType]]:
        """Return a list of numpy getter functions for all ROIs."""
        yield from (self.build_numpy_getter(roi) for roi in self.rois)

    def _dask_getters_generator(self) -> Generator[DataGetterProtocol[DaskPipeType]]:
        """Return a list of dask getter functions for all ROIs."""
        yield from (self.build_dask_getter(roi) for roi in self.rois)

    def _numpy_setters_generator(
        self,
    ) -> Generator[DataSetterProtocol[NumpyPipeType] | None]:
        """Return a list of numpy setter functions for all ROIs."""
        yield from (self.build_numpy_setter(roi) for roi in self.rois)

    def _dask_setters_generator(
        self,
    ) -> Generator[DataSetterProtocol[DaskPipeType] | None]:
        """Return a list of dask setter functions for all ROIs."""
        yield from (self.build_dask_setter(roi) for roi in self.rois)

    def _read_and_write_generator(
        self,
        getters: Generator[
            DataGetterProtocol[NumpyPipeType] | DataGetterProtocol[DaskPipeType]
        ],
        setters: Generator[
            DataSetterProtocol[NumpyPipeType] | DataSetterProtocol[DaskPipeType] | None
        ],
    ) -> Generator[
        tuple[
            DataGetterProtocol[NumpyPipeType] | DataGetterProtocol[DaskPipeType],
            DataSetterProtocol[NumpyPipeType] | DataSetterProtocol[DaskPipeType],
        ]
    ]:
        """Create an iterator over the pixels of the ROIs."""
        for getter, setter in zip(getters, setters, strict=True):
            if setter is None:
                name = self.__class__.__name__
                raise NgioValueError(f"Iterator is read-only: {name}")
            yield getter, setter
        self.post_consolidate()

    @overload
    def iter(
        self,
        lazy: Literal[True],
        data_mode: Literal["numpy"],
        iterator_mode: Literal["readwrite"],
    ) -> Generator[
        tuple[DataGetterProtocol[NumpyPipeType], DataSetterProtocol[NumpyPipeType]]
    ]: ...

    @overload
    def iter(
        self,
        lazy: Literal[True],
        data_mode: Literal["numpy"],
        iterator_mode: Literal["readonly"] = ...,
    ) -> Generator[DataGetterProtocol[NumpyPipeType]]: ...

    @overload
    def iter(
        self,
        lazy: Literal[True],
        data_mode: Literal["dask"],
        iterator_mode: Literal["readwrite"],
    ) -> Generator[
        tuple[DataGetterProtocol[DaskPipeType], DataSetterProtocol[DaskPipeType]]
    ]: ...

    @overload
    def iter(
        self,
        lazy: Literal[True],
        data_mode: Literal["dask"],
        iterator_mode: Literal["readonly"] = ...,
    ) -> Generator[DataGetterProtocol[DaskPipeType]]: ...

    @overload
    def iter(
        self,
        lazy: Literal[False],
        data_mode: Literal["numpy"],
        iterator_mode: Literal["readwrite"],
    ) -> Generator[tuple[NumpyPipeType, DataSetterProtocol[NumpyPipeType]]]: ...

    @overload
    def iter(
        self,
        lazy: Literal[False],
        data_mode: Literal["numpy"],
        iterator_mode: Literal["readonly"] = ...,
    ) -> Generator[NumpyPipeType]: ...

    @overload
    def iter(
        self,
        lazy: Literal[False],
        data_mode: Literal["dask"],
        iterator_mode: Literal["readwrite"],
    ) -> Generator[tuple[DaskPipeType, DataSetterProtocol[DaskPipeType]]]: ...

    @overload
    def iter(
        self,
        lazy: Literal[False],
        data_mode: Literal["dask"],
        iterator_mode: Literal["readonly"] = ...,
    ) -> Generator[DaskPipeType]: ...

    def iter(
        self,
        lazy: bool = False,
        data_mode: Literal["numpy", "dask"] = "dask",
        iterator_mode: Literal["readwrite", "readonly"] = "readwrite",
    ) -> Generator:
        """Create an iterator over the pixels of the ROIs."""
        if data_mode == "numpy":
            getters = self._numpy_getters_generator()
            setters = self._numpy_setters_generator()
        elif data_mode == "dask":
            getters = self._dask_getters_generator()
            setters = self._dask_setters_generator()
        else:
            raise NgioValueError(f"Invalid mode: {data_mode}")

        if iterator_mode == "readonly":
            if lazy:
                return getters
            else:
                return (getter() for getter in getters)
        if lazy:
            return self._read_and_write_generator(getters, setters)
        else:
            gen = self._read_and_write_generator(getters, setters)
            return ((getter(), setter) for getter, setter in gen)

    def iter_as_numpy(
        self,
    ):
        """Create an iterator over the pixels of the ROIs."""
        return self.iter(lazy=False, data_mode="numpy", iterator_mode="readwrite")

    def iter_as_dask(
        self,
    ):
        """Create an iterator over the pixels of the ROIs."""
        return self.iter(lazy=False, data_mode="dask", iterator_mode="readwrite")

    def map_as_numpy(
        self,
        func: Callable[[NumpyPipeType], NumpyPipeType],
        mapper: MapperProtocol[NumpyPipeType] | None = None,
    ) -> None:
        """Apply a transformation function to the ROI pixels."""
        if mapper is None:
            _mapper = BasicMapper[NumpyPipeType]()
        else:
            _mapper = mapper

        _mapper(
            func=func,
            getters=self._numpy_getters_generator(),
            setters=self._numpy_setters_generator(),
        )
        self.post_consolidate()

    def map_as_dask(
        self,
        func: Callable[[DaskPipeType], DaskPipeType],
        mapper: MapperProtocol[DaskPipeType] | None = None,
    ) -> None:
        """Apply a transformation function to the ROI pixels."""
        if mapper is None:
            _mapper = BasicMapper[DaskPipeType]()
        else:
            _mapper = mapper

        _mapper(
            func=func,
            getters=self._dask_getters_generator(),
            setters=self._dask_setters_generator(),
        )
        self.post_consolidate()

    def check_if_regions_overlap(self) -> bool:
        """Check if any of the ROIs overlap logically.

        If two ROIs cover the same pixel, they are considered to overlap.
        This does not consider chunking or other storage details.

        Returns:
            bool: True if any ROIs overlap. False otherwise.
        """
        if len(self.rois) < 2:
            # Less than 2 ROIs cannot overlap
            return False

        slicing_tuples = (
            g.slicing_ops.normalized_slicing_tuple
            for g in self._numpy_getters_generator()
        )
        x = check_if_regions_overlap(slicing_tuples)
        return x

    def require_no_regions_overlap(self) -> None:
        """Ensure that the Iterator's ROIs do not overlap."""
        if self.check_if_regions_overlap():
            raise NgioValueError("Some rois overlap.")

    def check_if_chunks_overlap(self) -> bool:
        """Check if any of the ROIs overlap in terms of chunks.

        If two ROIs cover the same chunk, they are considered to overlap in chunks.
        This does not consider pixel-level overlaps.

        Returns:
            bool: True if any ROIs overlap in chunks, False otherwise.
        """
        from ngio.io_pipes._ops_slices_utils import check_if_chunks_overlap

        if len(self.rois) < 2:
            # Less than 2 ROIs cannot overlap
            return False

        slicing_tuples = (
            g.slicing_ops.normalized_slicing_tuple
            for g in self._numpy_getters_generator()
        )
        shape = self.ref_image.shape
        chunks = self.ref_image.chunks
        return check_if_chunks_overlap(slicing_tuples, shape, chunks)

    def require_no_chunks_overlap(self) -> None:
        """Ensure that the ROIs do not overlap in terms of chunks."""
        if self.check_if_chunks_overlap():
            raise NgioValueError("Some rois overlap in chunks.")



import pytest

from ngio import PixelSize
from ngio.common import Roi
from ngio.utils import NgioValueError


def test_basic_rois_ops():
    roi = Roi(
        name="test",
        x=0.0,
        y=0.0,
        z=0.0,
        x_length=1.0,
        y_length=1.0,
        z_length=1.0,
        label=1,
        unit="micrometer",  # type: ignore
        other="other",  # type: ignore
    )

    assert roi.x == 0.0

    pixel_size = PixelSize(x=1.0, y=1.0, z=1.0)
    raster_roi = roi.to_roi_pixels(pixel_size)
    assert roi.__str__()
    assert roi.__repr__()

    assert raster_roi.to_slicing_dict(pixel_size=pixel_size) == {
        "x": slice(0, 1),
        "y": slice(0, 1),
        "z": slice(0, 1),
        "t": slice(None),
    }
    assert roi.model_extra is not None
    assert roi.model_extra["other"] == "other"

    world_roi_2 = raster_roi.to_roi(pixel_size)

    assert world_roi_2.x == 0.0
    assert world_roi_2.y == 0.0
    assert world_roi_2.z == 0.0
    assert world_roi_2.x_length == 1.0
    assert world_roi_2.y_length == 1.0
    assert world_roi_2.z_length == 1.0
    assert world_roi_2.other == "other"  # type: ignore

    roi_zoomed = roi.zoom(2.0)
    with pytest.raises(ValueError):
        roi.zoom(-1.0)

    assert roi_zoomed.to_slicing_dict(pixel_size) == {
        "x": slice(0, 2),
        "y": slice(0, 2),
        "z": slice(0, 1),
        "t": slice(None),
    }

    roi2 = Roi(
        name="test2",
        x=0.0,
        y=0.0,
        z=0.0,
        x_length=1.0,
        y_length=1.0,
        z_length=1.0,
        unit="micrometer",  # type: ignore
        label=1,
    )
    roi_i = roi.intersection(roi2)
    assert roi_i is not None
    assert roi_i.label == 1

    roi2.label = 2
    with pytest.raises(NgioValueError):
        roi.intersection(roi2)


@pytest.mark.parametrize(
    "roi_ref,roi_other,expected_intersection,expected_name",
    [
        (
            # Basic intersection
            Roi(
                name="ref",
                x=0.0,
                y=0.0,
                z=0.0,
                x_length=1.0,
                y_length=1.0,
                z_length=1.0,
                unit="micrometer",  # type: ignore
            ),
            Roi(
                name="other",
                x=0.5,
                y=0.5,
                z=0.5,
                x_length=1.0,
                y_length=1.0,
                z_length=1.0,
                unit="micrometer",  # type: ignore
            ),
            Roi(
                name="ref:other",
                x=0.5,
                y=0.5,
                z=0.5,
                x_length=0.5,
                y_length=0.5,
                z_length=0.5,
                unit="micrometer",  # type: ignore
            ),
            "ref:other",
        ),
        (
            # No intersection
            Roi(
                name="ref",
                x=0.0,
                y=0.0,
                z=0.0,
                x_length=1.0,
                y_length=1.0,
                z_length=1.0,
                unit="micrometer",  # type: ignore
            ),
            Roi(
                name="other",
                x=2.0,
                y=2.0,
                z=2.0,
                x_length=1.0,
                y_length=1.0,
                z_length=1.0,
                unit="micrometer",  # type: ignore
            ),
            None,
            "",
        ),
        (
            # Intersection with z=None (expected behaves like infinite z)
            # t=None (expected behaves like infinite t)
            Roi(
                name="ref",
                x=0.0,
                y=0.0,
                z=None,
                t=0,
                x_length=2.0,
                y_length=2.0,
                z_length=None,
                t_length=2.0,
                unit="micrometer",  # type: ignore
            ),
            Roi(
                name=None,
                x=-1.0,
                y=-1.0,
                z=-1.0,
                t=None,
                x_length=2.0,
                y_length=2.0,
                z_length=2.0,
                t_length=None,
                unit="micrometer",  # type: ignore
            ),
            Roi(
                name="ref",
                x=0.0,
                y=0.0,
                z=-1.0,
                t=0,
                x_length=1.0,
                y_length=1.0,
                z_length=2.0,
                t_length=2.0,
                unit="micrometer",  # type: ignore
            ),
            "ref",
        ),
    ],
)
def test_rois_intersection(
    roi_ref: Roi,
    roi_other: Roi,
    expected_intersection: Roi | None,
    expected_name: str,
):
    intersection = roi_ref.intersection(roi_other)
    if expected_intersection is None:
        assert intersection is None
    else:
        assert intersection is not None
        assert intersection.name == expected_name
        assert intersection.x == expected_intersection.x
        assert intersection.y == expected_intersection.y
        assert intersection.z == expected_intersection.z
        assert intersection.x_length == expected_intersection.x_length
        assert intersection.y_length == expected_intersection.y_length
        assert intersection.z_length == expected_intersection.z_length



"""HCS (High Content Screening) specific metadata classes for NGIO."""

from typing import Annotated

from ome_zarr_models.v04.hcs import HCSAttrs
from ome_zarr_models.v04.plate import (
    Acquisition,
    Column,
    Plate,
    Row,
    WellInPlate,
)
from ome_zarr_models.v04.well import WellAttrs as WellAttrs04
from ome_zarr_models.v04.well_types import WellImage as WellImage04
from ome_zarr_models.v04.well_types import WellMeta as WellMeta04
from pydantic import BaseModel, SkipValidation, field_serializer

from ngio.ome_zarr_meta.ngio_specs._ngio_image import DefaultNgffVersion, NgffVersions
from ngio.utils import NgioValueError, ngio_logger


def path_in_well_validation(path: str) -> str:
    """Validate the path in the well."""
    # Check if the value contains only alphanumeric characters
    if not path.isalnum():
        ngio_logger.warning(
            f"Path '{path}' contains non-alphanumeric characters. "
            "This may cause issues with some tools. "
            "Consider using only alphanumeric characters in the path."
        )
    return path


class ImageInWellPath(BaseModel):
    """Image in a well."""

    row: str
    column: str | int
    path: str
    acquisition_id: int | None = None
    acquisition_name: str | None = None


class CustomWellImage(WellImage04):
    path: Annotated[str, SkipValidation]

    @field_serializer("path")
    def serialize_path(self, value: str) -> str:
        """Custom serialization for the path."""
        return path_in_well_validation(value)


class CustomWellMeta(WellMeta04):
    images: list[CustomWellImage]  # type: ignore (override of WellMeta04.images)


class CustomWellAttrs(WellAttrs04):
    well: CustomWellMeta  # type: ignore (override of WellAttrs04.well)


class NgioWellMeta(CustomWellAttrs):
    """HCS well metadata."""

    @classmethod
    def default_init(
        cls,
        version: NgffVersions | None = None,
    ) -> "NgioWellMeta":
        if version is None:
            version = DefaultNgffVersion
        well = cls(well=CustomWellMeta(images=[], version=version))
        return well

    @property
    def acquisition_ids(self) -> list[int]:
        """Return the acquisition ids in the well."""
        acquisitions = []
        for images in self.well.images:
            if (
                images.acquisition is not None
                and images.acquisition not in acquisitions
            ):
                acquisitions.append(images.acquisition)
        return acquisitions

    def get_image_acquisition_id(self, image_path: str) -> int | None:
        """Return the acquisition id for the given image path."""
        for images in self.well.images:
            if images.path == image_path:
                return images.acquisition
        raise NgioValueError(f"Image at path {image_path} not found in the well.")

    def paths(self, acquisition: int | None = None) -> list[str]:
        """Return the images paths in the well.

        If acquisition is None, return all images paths in the well.
        Else, return the images paths in the well for the given acquisition.

        Args:
            acquisition (int | None): The acquisition id to filter the images.
        """
        if acquisition is None:
            return [images.path for images in self.well.images]
        return [
            images.path
            for images in self.well.images
            if images.acquisition == acquisition
        ]

    def add_image(
        self, path: str, acquisition: int | None = None, strict: bool = True
    ) -> "NgioWellMeta":
        """Add an image to the well.

        Args:
            path (str): The path of the image.
            acquisition (int | None): The acquisition id of the image.
            strict (bool): If True, check if the image already exists in the well.
                If False, do not check if the image already exists in the well.
        """
        list_of_images = self.well.images
        for image in list_of_images:
            if image.path == path:
                raise NgioValueError(
                    f"Image at path {path} already exists in the well."
                )

        if (
            strict
            and (acquisition is not None)
            and (acquisition not in self.acquisition_ids)
        ):
            raise NgioValueError(
                f"Acquisition ID {acquisition} not found in well. "
                "Please add it to the plate metadata first."
            )

        new_image = CustomWellImage(path=path, acquisition=acquisition)
        list_of_images.append(new_image)
        return NgioWellMeta(
            well=CustomWellMeta(images=list_of_images, version=self.well.version)
        )

    def remove_image(self, path: str) -> "NgioWellMeta":
        """Remove an image from the well.

        Args:
            path (str): The path of the image.
        """
        list_of_images = self.well.images
        for image in list_of_images:
            if image.path == path:
                list_of_images.remove(image)
                return NgioWellMeta(
                    well=CustomWellMeta(
                        images=list_of_images, version=self.well.version
                    )
                )
        raise NgioValueError(f"Image at path {path} not found in the well.")


def _stringify_column(column: str | int) -> str:
    """Convert the column to a string.

    Args:
        column (str | int): The column to convert.

    Returns:
        str: The column as a string.
    """
    if isinstance(column, str):
        return column

    # Maybe we should pad the column with zeros
    return str(column)


def _find_row_index(rows: list[str], row: str) -> int | None:
    try:
        return rows.index(row)
    except ValueError:
        return None


def _find_column_index(columns: list[str], column: str | int) -> int | None:
    _num_columns = [int(columns) for columns in columns]
    column = int(column)
    try:
        return _num_columns.index(column)
    except ValueError:
        return None


def _relabel_wells(
    wells: list[WellInPlate], rows: list[Row], columns: list[Column]
) -> list[WellInPlate]:
    new_wells = []
    _rows = [row.name for row in rows]
    _columns = [column.name for column in columns]
    for well in wells:
        row, column = well.path.split("/")
        row_idx = _find_row_index(_rows, row)
        column_idx = _find_column_index(_columns, column)

        if row_idx is None:
            raise NgioValueError(f"Row {row} not found in the plate.")
        if column_idx is None:
            raise NgioValueError(f"Column {column} not found in the plate.")

        new_wells.append(
            WellInPlate(
                path=well.path,
                rowIndex=row_idx,
                columnIndex=column_idx,
            )
        )

    return new_wells


class NgioPlateMeta(HCSAttrs):
    """HCS plate metadata."""

    @classmethod
    def default_init(
        cls,
        images: list[ImageInWellPath] | None = None,
        name: str | None = None,
        version: NgffVersions | None = None,
    ) -> "NgioPlateMeta":
        plate = cls(
            plate=Plate(
                rows=[],
                columns=[],
                acquisitions=None,
                wells=[],
                field_count=None,
                version=version,
                name=name,
            )
        )

        if images is None:
            return plate

        for image in images:
            plate = plate.add_well(
                row=image.row,
                column=image.column,
            )
            if image.acquisition_id is not None:
                plate = plate.add_acquisition(
                    acquisition_id=image.acquisition_id,
                    acquisition_name=image.acquisition_name,
                )
        return plate

    @property
    def columns(self) -> list[str]:
        """Returns the list of columns in the plate."""
        return [columns.name for columns in self.plate.columns]

    @property
    def rows(self) -> list[str]:
        """Returns the list of rows in the plate."""
        return [rows.name for rows in self.plate.rows]

    @property
    def acquisitions_names(self) -> list[str | None]:
        """Return the acquisitions in the plate."""
        if self.plate.acquisitions is None:
            return []
        return [acquisitions.name for acquisitions in self.plate.acquisitions]

    @property
    def acquisition_ids(self) -> list[int]:
        """Return the acquisitions ids in the plate."""
        if self.plate.acquisitions is None:
            return []
        return [acquisitions.id for acquisitions in self.plate.acquisitions]

    @property
    def wells_paths(self) -> list[str]:
        """Return the wells paths in the plate."""
        return [wells.path for wells in self.plate.wells]

    def get_well_path(self, row: str, column: str | int) -> str:
        """Return the well path for the given row and column.

        Args:
            row (str): The row of the well.
            column (str | int): The column of the well.

        Returns:
            str: The path of the well.
        """
        if row not in self.rows:
            raise NgioValueError(
                f"Row {row} not found in the plate. Available rows are {self.rows}."
            )

        row_idx = self.rows.index(row)

        _num_columns = [int(columns) for columns in self.columns]

        try:
            _column = int(column)
        except ValueError:
            raise NgioValueError(
                f"Column {column} must be an integer or convertible to an integer."
            ) from None

        column_idx = _num_columns.index(_column)

        for well in self.plate.wells:
            if well.columnIndex == column_idx and well.rowIndex == row_idx:
                return well.path

        raise NgioValueError(
            f"Well at row {row} and column {column} not found in the plate."
        )

    def add_row(self, row: str) -> "tuple[NgioPlateMeta, int]":
        """Add a row to the plate.

        Args:
            row (str): The row to add.
        """
        relabel_wells = False

        row_names = self.rows
        row_idx = _find_row_index(row_names, row)
        if row_idx is not None:
            # Nothing to do
            return self, row_idx

        row_names.append(row)
        row_names.sort()
        row_idx = row_names.index(row)
        relabel_wells = True

        rows = [Row(name=row) for row in row_names]

        if relabel_wells:
            wells = _relabel_wells(self.plate.wells, rows, self.plate.columns)
        else:
            wells = self.plate.wells

        new_plate = Plate(
            rows=rows,
            columns=self.plate.columns,
            acquisitions=self.plate.acquisitions,
            wells=wells,
            field_count=self.plate.field_count,
            name=self.plate.name,
            version=self.plate.version,
        )
        return NgioPlateMeta(plate=new_plate), row_idx

    def add_column(self, column: str | int) -> "tuple[NgioPlateMeta, int]":
        """Add a column to the plate.

        Args:
            column (str | int): The column to add.
        """
        relabel_wells = False

        columns_names = self.columns
        column_idx = _find_column_index(columns_names, column)
        if column_idx is not None:
            # Nothing to do
            return self, column_idx

        columns_names.append(_stringify_column(column))
        # sort as numbers
        columns_names.sort(key=lambda x: int(x))
        column_idx = columns_names.index(_stringify_column(column))
        relabel_wells = True

        columns = [Column(name=column) for column in columns_names]

        if relabel_wells:
            wells = _relabel_wells(self.plate.wells, self.plate.rows, columns)
        else:
            wells = self.plate.wells

        new_plate = Plate(
            rows=self.plate.rows,
            columns=columns,
            acquisitions=self.plate.acquisitions,
            wells=wells,
            field_count=self.plate.field_count,
            name=self.plate.name,
            version=self.plate.version,
        )
        return NgioPlateMeta(plate=new_plate), column_idx

    def add_well(
        self,
        row: str,
        column: str | int,
    ) -> "NgioPlateMeta":
        """Add an image to the well.

        Args:
            row (str): The row of the well.
            column (str | int): The column of the well.
        """
        plate, row_idx = self.add_row(row=row)
        plate, column_idx = plate.add_column(column=column)

        wells = plate.plate.wells
        for well_obj in wells:
            if well_obj.rowIndex == row_idx and well_obj.columnIndex == column_idx:
                break
        else:
            wells.append(
                WellInPlate(
                    path=f"{row}/{_stringify_column(column)}",
                    rowIndex=row_idx,
                    columnIndex=column_idx,
                )
            )

        new_plate = Plate(
            rows=plate.plate.rows,
            columns=plate.plate.columns,
            acquisitions=plate.plate.acquisitions,
            wells=wells,
            field_count=plate.plate.field_count,
            name=plate.plate.name,
            version=plate.plate.version,
        )
        return NgioPlateMeta(plate=new_plate)

    def add_acquisition(
        self,
        acquisition_id: int,
        acquisition_name: str | None = None,
        **acquisition_kwargs,
    ) -> "NgioPlateMeta":
        """Add an acquisition to the plate.

        Args:
            acquisition_id (int): The acquisition id of the well.
            acquisition_name (str | None): The acquisition name of the well.
            **acquisition_kwargs: Additional acquisition metadata.
        """
        acquisitions = self.plate.acquisitions
        if acquisitions is None:
            acquisitions = []

        for acquisition_obj in acquisitions:
            if acquisition_obj.id == acquisition_id:
                # If the acquisition already exists
                # Nothing to do
                # Maybe we should update the acquisition name and kwargs
                return self

        acquisitions.append(
            Acquisition(id=acquisition_id, name=acquisition_name, **acquisition_kwargs)
        )

        new_plate = Plate(
            rows=self.plate.rows,
            columns=self.plate.columns,
            acquisitions=acquisitions,
            wells=self.plate.wells,
            field_count=self.plate.field_count,
            name=self.plate.name,
            version=self.plate.version,
        )
        return NgioPlateMeta(plate=new_plate)

    def remove_well(self, row: str, column: str | int) -> "NgioPlateMeta":
        """Remove a well from the plate.

        Args:
            row (str): The row of the well.
            column (str | int): The column of the well.
        """
        row_idx = _find_row_index(self.rows, row)
        if row_idx is None:
            raise NgioValueError(f"Row {row} not found in the plate.")

        column_idx = _find_column_index(self.columns, column)
        if column_idx is None:
            raise NgioValueError(f"Column {column} not found in the plate.")

        wells = self.plate.wells
        for well_obj in wells:
            if well_obj.rowIndex == row_idx and well_obj.columnIndex == column_idx:
                wells.remove(well_obj)
                break
        else:
            raise NgioValueError(
                f"Well at row {row} and column {column} not found in the plate."
            )

        new_plate = Plate(
            rows=self.plate.rows,
            columns=self.plate.columns,
            acquisitions=self.plate.acquisitions,
            wells=wells,
            field_count=self.plate.field_count,
            name=self.plate.name,
            version=self.plate.version,
        )
        return NgioPlateMeta(plate=new_plate)

    def derive(
        self,
        name: str | None = None,
        version: NgffVersions | None = None,
        keep_acquisitions: bool = False,
    ) -> "NgioPlateMeta":
        """Derive the plate metadata.

        Args:
            name (str): The name of the derived plate.
            version (NgffVersion | None): The version of the derived plate.
                If None, use the version of the original plate.
            keep_acquisitions (bool): If True, keep the acquisitions in the plate.
        """
        columns = self.plate.columns
        rows = self.plate.rows

        if keep_acquisitions:
            acquisitions = self.plate.acquisitions
        else:
            acquisitions = None

        if version is None:
            version = self.plate.version  # type: ignore (version is NgffVersions or None)

        return NgioPlateMeta(
            plate=Plate(
                rows=rows,
                columns=columns,
                acquisitions=acquisitions,
                wells=[],
                field_count=self.plate.field_count,
                version=version,
                name=name,
            )
        )



import asyncio
from pathlib import Path

import pandas.testing as pdt
import pytest
import zarr
from pandas import DataFrame

from ngio import (
    ImageInWellPath,
    OmeZarrContainer,
    OmeZarrWell,
    Roi,
    create_empty_plate,
    open_ome_zarr_plate,
)
from ngio.tables import GenericTable, RoiTable
from ngio.utils import NgioValueError


def test_open_real_ome_zarr_plate(cardiomyocyte_tiny_path: Path):
    cardiomyocyte_tiny_path = cardiomyocyte_tiny_path
    ome_zarr_plate = open_ome_zarr_plate(cardiomyocyte_tiny_path)

    assert isinstance(ome_zarr_plate.__repr__(), str)
    assert ome_zarr_plate.columns == ["03"]
    assert ome_zarr_plate.rows == ["B"]
    assert ome_zarr_plate.acquisition_ids == [0]
    assert ome_zarr_plate.acquisitions_names == [
        "20200812-CardiomyocyteDifferentiation14-Cycle1"
    ]

    well_path = ome_zarr_plate._well_path("B", "03")
    well_path2 = ome_zarr_plate._well_path("B", 3)
    assert well_path == well_path2
    well = ome_zarr_plate.get_well("B", "03")
    assert well.paths() == ["0"]

    image_path = ome_zarr_plate._image_path("B", "03", "0")
    assert image_path == "B/03/0"

    images_plate = ome_zarr_plate.get_images()
    assert len(images_plate) == 1
    images_plate_async = asyncio.run(ome_zarr_plate.get_images_async())
    assert len(images_plate_async) == 1
    assert images_plate.keys() == images_plate_async.keys()

    _ = ome_zarr_plate.get_image("B", "03", "0")

    images_well = ome_zarr_plate.get_well_images("B", "03")
    assert len(images_well) == 1

    well = ome_zarr_plate.get_well("B", "03")
    assert isinstance(well, OmeZarrWell)
    assert well.paths() == ["0"]

    image = well.get_image("0")
    assert isinstance(image, OmeZarrContainer)
    assert well.get_image_acquisition_id("0") is None


def test_create_and_edit_plate(tmp_path: Path):
    test_plate = create_empty_plate(tmp_path / "test_plate.zarr", name="test_plate")
    assert test_plate.columns == []
    assert test_plate.rows == []
    assert test_plate.acquisition_ids == []

    test_plate.add_image(row="B", column="03", image_path="0", acquisition_id=0)
    test_plate.add_image(row="B", column="03", image_path="1", acquisition_id=0)

    assert test_plate.meta.plate.name == "test_plate"

    with pytest.raises(NgioValueError):
        test_plate.add_image(row="B", column="03", image_path="1", acquisition_id=1)

    test_plate.atomic_add_image(row="C", column="02", image_path="1", acquisition_id=1)

    assert test_plate.columns == ["02", "03"]
    assert test_plate.rows == ["B", "C"]
    assert test_plate.acquisition_ids == [0, 1]
    assert (
        test_plate.get_image_acquisition_id(row="B", column="03", image_path="0") == 0
    )

    assert len(test_plate.wells_paths()) == 2

    test_plate.remove_image(row="C", column="02", image_path="1")
    assert len(test_plate.wells_paths()) == 1


def test_create_and_edit_plate_path_normalization(tmp_path: Path):
    test_plate = create_empty_plate(tmp_path / "test_plate.zarr", name="test_plate")
    test_plate.add_image(row="B", column="03", image_path="0_mip", acquisition_id=0)
    test_plate.add_image(
        row="B", column="03", image_path="1_illumination_correction", acquisition_id=0
    )
    assert test_plate.images_paths() == ["B/03/0_mip", "B/03/1_illumination_correction"]


def test_derive_plate_from_ome_zarr(cardiomyocyte_tiny_path: Path, tmp_path: Path):
    ome_zarr_plate = open_ome_zarr_plate(cardiomyocyte_tiny_path)
    test_plate = ome_zarr_plate.derive_plate(
        tmp_path / "test_plate.zarr", keep_acquisitions=True
    )
    assert test_plate.columns == ["03"]
    assert test_plate.rows == ["B"]
    assert test_plate.acquisition_ids == [0]


def test_add_well(tmp_path: Path):
    test_plate = create_empty_plate(tmp_path / "test_plate.zarr", name="test_plate")
    well = test_plate.add_well(row="B", column="03")
    assert isinstance(well, OmeZarrWell)
    assert test_plate.columns == ["03"]
    assert test_plate.rows == ["B"]
    assert test_plate.acquisition_ids == []
    assert test_plate.wells_paths() == ["B/03"]
    assert test_plate.meta.plate.name == "test_plate"

    test_plate.add_column("04")
    test_plate.add_row("C")
    assert test_plate.columns == ["03", "04"]
    assert test_plate.rows == ["B", "C"]
    # No well added in this step
    assert test_plate.wells_paths() == ["B/03"]


def test_add_well_with_acquisition(tmp_path: Path):
    test_plate = create_empty_plate(tmp_path / "test_plate.zarr", name="test_plate")
    test_plate.add_acquisition(acquisition_id=0, acquisition_name="test_acquisition")
    test_plate.add_acquisition(acquisition_id=1, acquisition_name="test_acquisition1")
    assert test_plate.acquisition_ids == [0, 1]
    assert test_plate.acquisitions_names == ["test_acquisition", "test_acquisition1"]


def test_create_plate_with_wells(tmp_path: Path):
    images = [
        ImageInWellPath(row="B", column="03", path="0", acquisition_id=0),
        ImageInWellPath(row="B", column="03", path="1", acquisition_id=1),
        ImageInWellPath(row="C", column="02", path="0"),
    ]

    test_plate = create_empty_plate(
        tmp_path / "test_plate.zarr", name="test_plate", images=images
    )
    assert test_plate.columns == ["02", "03"]
    assert test_plate.rows == ["B", "C"]
    assert test_plate.acquisition_ids == [0, 1]
    assert test_plate.wells_paths() == ["B/03", "C/02"]
    assert test_plate.images_paths() == ["B/03/0", "B/03/1", "C/02/0"]
    assert test_plate.get_well("B", "03").acquisition_ids == [0, 1]
    assert test_plate.get_well("C", "02").acquisition_ids == []
    well = test_plate.get_well("B", "03")
    assert isinstance(well, OmeZarrWell)
    assert well.paths() == ["0", "1"]
    assert well.get_image_acquisition_id("0") == 0
    assert well.get_image_acquisition_id("1") == 1

    store = test_plate.get_image_store(row="B", column="03", image_path="0")
    assert isinstance(store, zarr.Group)


def test_tables_api(tmp_path: Path):
    test_plate = create_empty_plate(tmp_path / "test_plate.zarr", name="test_plate")

    test_df = DataFrame({"a": [1, 2], "b": [3, 4]})
    test_table = GenericTable(test_df)
    test_plate.add_table("test_table", test_table, backend="csv")

    test_roi_table = RoiTable(
        rois=[Roi(name="roi_1", x_length=10, y_length=10, z_length=10)]  # type: ignore
    )
    test_plate.add_table("test_roi_table", test_roi_table)
    assert test_plate.list_tables() == ["test_table", "test_roi_table"]
    assert test_plate.list_roi_tables() == ["test_roi_table"]

    pdt.assert_frame_equal(
        test_plate.get_table("test_table").dataframe,  # type: ignore
        test_df,
        check_names=False,
    )


@pytest.mark.filterwarnings("ignore::anndata._warnings.ImplicitModificationWarning")
def test_plate_table_aggregations(cardiomyocyte_small_mip_path: Path):
    ome_zarr_plate = open_ome_zarr_plate(cardiomyocyte_small_mip_path)
    expected_tables = [
        "FOV_ROI_table",
        "nuclei_ROI_table",
        "well_ROI_table",
        "regionprops_DAPI",
        "nuclei_measurements_wf3",
        "nuclei_measurements_wf4",
        "nuclei_lamin_measurements_wf4",
    ]
    tables = ome_zarr_plate.list_image_tables()
    assert set(tables) == set(expected_tables)
    async_tables = asyncio.run(ome_zarr_plate.list_image_tables_async())
    assert set(async_tables) == set(expected_tables)
    roi_tables = ome_zarr_plate.list_image_tables(filter_types="roi_table")
    assert set(roi_tables) == {"FOV_ROI_table", "well_ROI_table"}

    t1 = ome_zarr_plate.concatenate_image_tables(name="regionprops_DAPI")
    t2 = asyncio.run(
        ome_zarr_plate.concatenate_image_tables_async(name="regionprops_DAPI")
    )
    pdt.assert_frame_equal(t1.dataframe, t2.dataframe)



"""Implementation of a generic table class."""

from ngio.tables._abstract_table import AbstractBaseTable
from ngio.tables.backends import (
    BackendMeta,
    TableBackend,
    TabularData,
)
from ngio.utils import ZarrGroupHandler


class ConditionTableMeta(BackendMeta):
    """Metadata for the condition table."""

    table_version: str | None = "1"
    type: str | None = "condition_table"


class ConditionTableV1(AbstractBaseTable):
    """Condition table class.

    This class is used to load a condition table.
    The condition table is a generic table that does not
    have a specific definition.

    It is used to store information about the particular conditions
    used to generate the data.
    - How much drug was used in the experiment
    - What treatment was used
    - etc.
    """

    def __init__(
        self,
        table_data: TabularData | None = None,
        *,
        meta: ConditionTableMeta | None = None,
    ) -> None:
        """Initialize the ConditionTable."""
        if meta is None:
            meta = ConditionTableMeta()

        super().__init__(
            table_data=table_data,
            meta=meta,
        )

    @staticmethod
    def table_type() -> str:
        """Return the type of the table."""
        return "condition_table"

    @staticmethod
    def version() -> str:
        """The generic table does not have a version.

        Since does not follow a specific schema.
        """
        return "1"

    @classmethod
    def from_handler(
        cls,
        handler: ZarrGroupHandler,
        backend: TableBackend | None = None,
    ) -> "ConditionTableV1":
        return cls._from_handler(
            handler=handler,
            backend=backend,
            meta_model=ConditionTableMeta,
        )



"""Abstract class for handling OME-NGFF images."""

from collections.abc import Sequence
from typing import Literal

import numpy as np
import PIL.Image
from zarr.core.array import CompressorLike

from ngio.common._synt_images_utils import fit_to_shape
from ngio.images._ome_zarr_container import OmeZarrContainer, create_ome_zarr_from_array
from ngio.ome_zarr_meta.ngio_specs import (
    DefaultNgffVersion,
    NgffVersions,
)
from ngio.resources import AVAILABLE_SAMPLES, SampleInfo, get_sample_info
from ngio.tables import (
    DefaultTableBackend,
    TableBackend,
)
from ngio.utils import (
    StoreOrGroup,
)


def create_synthetic_ome_zarr(
    store: StoreOrGroup,
    shape: Sequence[int],
    reference_sample: AVAILABLE_SAMPLES | SampleInfo = "Cardiomyocyte",
    levels: int | list[str] = 5,
    xy_scaling_factor: float = 2,
    z_scaling_factor: float = 1.0,
    axes_names: Sequence[str] | None = None,
    chunks: Sequence[int] | Literal["auto"] = "auto",
    channel_labels: list[str] | None = None,
    channel_wavelengths: list[str] | None = None,
    channel_colors: Sequence[str] | None = None,
    channel_active: Sequence[bool] | None = None,
    table_backend: TableBackend = DefaultTableBackend,
    dimension_separator: Literal[".", "/"] = "/",
    compressors: CompressorLike = "auto",
    overwrite: bool = False,
    version: NgffVersions = DefaultNgffVersion,
) -> OmeZarrContainer:
    """Create an empty OME-Zarr image with the given shape and metadata.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        shape (Sequence[int]): The shape of the image.
        reference_sample (AVAILABLE_SAMPLES | SampleInfo): The reference sample to use.
        levels (int | list[str], optional): The number of levels in the pyramid or a
            list of level names. Defaults to 5.
        xy_scaling_factor (float, optional): The down-scaling factor in x and y
            dimensions. Defaults to 2.0.
        z_scaling_factor (float, optional): The down-scaling factor in z dimension.
            Defaults to 1.0.
        axes_names (Sequence[str] | None, optional): The names of the axes.
            If None the canonical names are used. Defaults to None.
        chunks (Sequence[int] | Literal["auto"]): The chunk shape. If None the shape
            is used. Defaults to "auto".
        channel_labels (list[str] | None, optional): The labels of the channels.
            Defaults to None.
        channel_wavelengths (list[str] | None, optional): The wavelengths of the
            channels. Defaults to None.
        channel_colors (Sequence[str] | None, optional): The colors of the channels.
            Defaults to None.
        channel_active (Sequence[bool] | None, optional): Whether the channels are
            active. Defaults to None.
        table_backend (TableBackend): Table backend to be used to store tables
        dimension_separator (DIMENSION_SEPARATOR): The separator to use for
            dimensions. Defaults to "/".
        compressors (CompressorLike): The compressors to use. Defaults to "auto".
        overwrite (bool, optional): Whether to overwrite an existing image.
            Defaults to True.
        version (NgffVersion, optional): The version of the OME-Zarr specification.
            Defaults to DefaultNgffVersion.
    """
    if isinstance(reference_sample, str):
        sample_info = get_sample_info(reference_sample)
    else:
        sample_info = reference_sample

    raw = np.asarray(PIL.Image.open(sample_info.img_path))
    raw = fit_to_shape(arr=raw, out_shape=tuple(shape))
    raw = raw / np.max(raw) * (2**16 - 1)
    raw = raw.astype(np.uint16)
    ome_zarr = create_ome_zarr_from_array(
        store=store,
        array=raw,
        xy_pixelsize=sample_info.xy_pixelsize,
        z_spacing=sample_info.z_spacing,
        time_spacing=sample_info.time_spacing,
        levels=levels,
        xy_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        space_unit=sample_info.space_unit,
        time_unit=sample_info.time_unit,
        axes_names=axes_names,
        channel_labels=channel_labels,
        channel_wavelengths=channel_wavelengths,
        channel_colors=channel_colors,
        channel_active=channel_active,
        name=sample_info.name,
        chunks=chunks,
        overwrite=overwrite,
        dimension_separator=dimension_separator,
        compressors=compressors,
        version=version,
    )

    image = ome_zarr.get_image()
    well_table = image.build_image_roi_table()
    ome_zarr.add_table("well_ROI_table", table=well_table, backend=table_backend)

    for label_info in sample_info.labels:
        ome_zarr.derive_label(name=label_info.name)
        label = ome_zarr.get_label(name=label_info.name)

        ref_label = np.asarray(PIL.Image.open(label_info.label_path))
        ref_label = ref_label.astype(label_info.dtype)

        ref_label = fit_to_shape(
            arr=ref_label,
            out_shape=label.shape,
            ensure_unique_info=label_info.ensure_unique_labels,
        )
        ref_label = ref_label.astype(np.uint32)
        label.set_array(ref_label)
        label.consolidate()

        if label_info.create_masking_table:
            masking_table = label.build_masking_roi_table()
            ome_zarr.add_table(
                name=f"{label_info.name}_masking_table",
                table=masking_table,
                backend=table_backend,
            )

    return ome_zarr



# Create a generic error class for the NGFF project


class NgioError(Exception):
    """Base class for all errors in the NGFF project."""

    pass


class NgioFileNotFoundError(NgioError, FileNotFoundError):
    """Error raised when a file is not found."""

    pass


class NgioFileExistsError(NgioError, FileExistsError):
    """Error raised when a file already exists."""

    pass


class NgioValidationError(NgioError, ValueError):
    """Generic error raised when a file does not pass validation."""

    pass


class NgioTableValidationError(NgioError):
    """Error raised when a table does not pass validation."""

    pass


class NgioValueError(NgioError, ValueError):
    """Error raised when a value does not pass a run time test."""

    pass



"""A module for handling the Plate Sequence in an OME-Zarr file."""

import asyncio
import warnings
from collections.abc import Sequence
from typing import Literal

from ngio.images import (
    OmeZarrContainer,
    concatenate_image_tables,
    concatenate_image_tables_as,
    concatenate_image_tables_as_async,
    concatenate_image_tables_async,
    list_image_tables,
    list_image_tables_async,
)
from ngio.ome_zarr_meta import (
    ImageInWellPath,
    NgffVersions,
    NgioPlateMeta,
    NgioWellMeta,
    find_plate_meta_handler,
    find_well_meta_handler,
    get_plate_meta_handler,
    get_well_meta_handler,
    path_in_well_validation,
)
from ngio.tables import (
    ConditionTable,
    DefaultTableBackend,
    FeatureTable,
    GenericRoiTable,
    MaskingRoiTable,
    RoiTable,
    Table,
    TableBackend,
    TablesContainer,
    TableType,
    TypedTable,
)
from ngio.utils import (
    AccessModeLiteral,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)


def _default_table_container(handler: ZarrGroupHandler) -> TablesContainer | None:
    """Return a default table container."""
    success, table_handler = handler.safe_derive_handler("tables")
    if success and isinstance(table_handler, ZarrGroupHandler):
        return TablesContainer(table_handler)


# Mock lock class that does nothing
class MockLock:
    """A mock lock class that does nothing."""

    def __enter__(self):
        """Enter the lock."""
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the lock."""
        pass


class OmeZarrWell:
    """A class to handle the Well Sequence in an OME-Zarr file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler.

        Args:
            group_handler: The Zarr group handler that contains the Well.
        """
        self._group_handler = group_handler
        self._meta_handler = find_well_meta_handler(group_handler)

    def __repr__(self) -> str:
        """Return a string representation of the well."""
        return f"Well(#images: {len(self.paths())})"

    @property
    def meta_handler(self):
        """Return the metadata handler."""
        return self._meta_handler

    @property
    def meta(self):
        """Return the metadata."""
        return self._meta_handler.meta

    @property
    def acquisition_ids(self) -> list[int]:
        """Return the acquisitions ids in the well."""
        return self.meta.acquisition_ids

    def paths(self, acquisition: int | None = None) -> list[str]:
        """Return the images paths in the well.

        If acquisition is None, return all images paths in the well.
        Else, return the images paths in the well for the given acquisition.

        Args:
            acquisition (int | None): The acquisition id to filter the images.
        """
        return self.meta.paths(acquisition)

    def get_image_store(self, image_path: str) -> StoreOrGroup:
        """Get the image store from the well.

        Args:
            image_path (str): The path of the image.
        """
        return self._group_handler.get_group(image_path, create_mode=True)

    def get_image_acquisition_id(self, image_path: str) -> int | None:
        """Get the acquisition id of an image in the well.

        Args:
            image_path (str): The path of the image.

        Returns:
            int | None: The acquisition id of the image.
        """
        return self.meta.get_image_acquisition_id(image_path=image_path)

    def get_image(self, image_path: str) -> OmeZarrContainer:
        """Get an image from the well.

        Args:
            image_path (str): The path of the image.

        Returns:
            OmeZarrContainer: The image.
        """
        handler = self._group_handler.derive_handler(image_path)
        return OmeZarrContainer(handler)

    def _add_image(
        self,
        image_path: str,
        acquisition_id: int | None = None,
        strict: bool = True,
        atomic: bool = False,
    ) -> StoreOrGroup:
        """Add an image to an ome-zarr well."""
        image_path = path_in_well_validation(path=image_path)

        if atomic:
            well_lock = self._group_handler.lock
        else:
            well_lock = MockLock()

        with well_lock:
            meta = self.meta.add_image(
                path=image_path, acquisition=acquisition_id, strict=strict
            )
            self.meta_handler.write_meta(meta)
            self.meta_handler._group_handler.clean_cache()

        return self._group_handler.get_group(image_path, create_mode=True)

    def atomic_add_image(
        self,
        image_path: str,
        acquisition_id: int | None = None,
        strict: bool = True,
    ) -> StoreOrGroup:
        """Parallel safe version of add_image."""
        return self._add_image(
            image_path=image_path,
            acquisition_id=acquisition_id,
            atomic=True,
            strict=strict,
        )

    def add_image(
        self,
        image_path: str,
        acquisition_id: int | None = None,
        strict: bool = True,
    ) -> StoreOrGroup:
        """Add an image to an ome-zarr well.

        Args:
            image_path (str): The path of the image.
            acquisition_id (int | None): The acquisition id to filter the images.
            strict (bool): Whether to check if the acquisition id is already exists
                in the well. Defaults to True. If False this might lead to
                acquisition in a well that does not exist at the plate level.
        """
        return self._add_image(
            image_path=image_path,
            acquisition_id=acquisition_id,
            atomic=False,
            strict=strict,
        )


def _build_extras(paths: Sequence[str]) -> list[dict[str, str]]:
    """Build the extras for the images.

    Args:
        paths (Sequence[str]): The paths of the images.

    Returns:
        list[dict[str, str]]: The extras for the images.
    """
    extras = []
    for path in paths:
        row, column, path_in_well = path.split("/")
        extras.append(
            {
                "row": row,
                "column": column,
                "path_in_well": path_in_well,
            }
        )
    return extras


class OmeZarrPlate:
    """A class to handle the Plate Sequence in an OME-Zarr file."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        table_container: TablesContainer | None = None,
    ) -> None:
        """Initialize the LabelGroupHandler.

        Args:
            group_handler: The Zarr group handler that contains the Plate.
            table_container: The tables container that contains plate level tables.
        """
        self._group_handler = group_handler
        self._meta_handler = find_plate_meta_handler(group_handler)
        self._tables_container = table_container

    def __repr__(self) -> str:
        """Return a string representation of the plate."""
        return f"Plate([rows x columns] ({len(self.rows)} x {len(self.columns)}))"

    @property
    def meta_handler(self):
        """Return the metadata handler."""
        return self._meta_handler

    @property
    def meta(self):
        """Return the metadata."""
        return self._meta_handler.meta

    @property
    def columns(self) -> list[str]:
        """Return the number of columns in the plate."""
        return self.meta.columns

    @property
    def rows(self) -> list[str]:
        """Return the number of rows in the plate."""
        return self.meta.rows

    @property
    def acquisitions_names(self) -> list[str | None]:
        """Return the acquisitions in the plate."""
        return self.meta.acquisitions_names

    @property
    def acquisition_ids(self) -> list[int]:
        """Return the acquisitions ids in the plate."""
        return self.meta.acquisition_ids

    def _well_path(self, row: str, column: int | str) -> str:
        """Return the well path in the plate."""
        return self.meta.get_well_path(row=row, column=column)

    def _image_path(self, row: str, column: int | str, path: str) -> str:
        """Return the image path in the plate."""
        well = self.get_well(row, column)
        if path not in well.paths():
            raise ValueError(f"Image {path} does not exist in well {row}{column}")
        return f"{self._well_path(row, column)}/{path}"

    def wells_paths(self) -> list[str]:
        """Return the wells paths in the plate."""
        return self.meta.wells_paths

    async def images_paths_async(self, acquisition: int | None = None) -> list[str]:
        """Return the images paths in the plate asynchronously.

        If acquisition is None, return all images paths in the plate.
        Else, return the images paths in the plate for the given acquisition.

        Args:
            acquisition (int | None): The acquisition id to filter the images.
        """
        wells = await self.get_wells_async()
        paths = []
        for well_path, well in wells.items():
            for img_path in well.paths(acquisition):
                paths.append(f"{well_path}/{img_path}")
        return paths

    def images_paths(self, acquisition: int | None = None) -> list[str]:
        """Return the images paths in the plate.

        If acquisition is None, return all images paths in the plate.
        Else, return the images paths in the plate for the given acquisition.

        Args:
            acquisition (int | None): The acquisition id to filter the images.
        """
        wells = self.get_wells()
        images = []
        for well_path, well in wells.items():
            for img_path in well.paths(acquisition):
                images.append(f"{well_path}/{img_path}")
        return images

    def well_images_paths(
        self, row: str, column: int | str, acquisition: int | None = None
    ) -> list[str]:
        """Return the images paths in a well.

        If acquisition is None, return all images paths in the well.
        Else, return the images paths in the well for the given acquisition.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.
            acquisition (int | None): The acquisition id to filter the images.
        """
        images = []
        well = self.get_well(row=row, column=column)
        for path in well.paths(acquisition):
            images.append(self._image_path(row=row, column=column, path=path))
        return images

    def get_image_acquisition_id(
        self, row: str, column: int | str, image_path: str
    ) -> int | None:
        """Get the acquisition id of an image in a well.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.
            image_path (str): The path of the image.

        Returns:
            int | None: The acquisition id of the image.
        """
        well = self.get_well(row=row, column=column)
        return well.get_image_acquisition_id(image_path=image_path)

    def get_well(self, row: str, column: int | str) -> OmeZarrWell:
        """Get a well from the plate.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.

        Returns:
            OmeZarrWell: The well.
        """
        well_path = self._well_path(row=row, column=column)
        group_handler = self._group_handler.derive_handler(well_path)
        return OmeZarrWell(group_handler)

    async def get_wells_async(self) -> dict[str, OmeZarrWell]:
        """Get all wells in the plate asynchronously.

        This method processes wells in parallel for improved performance
        when working with a large number of wells.

        Returns:
            dict[str, OmeZarrWell]: A dictionary of wells, where the key is the well
                path and the value is the well object.
        """
        wells = self._group_handler.get_from_cache("wells")
        if wells is not None:
            assert isinstance(wells, dict)
            return wells

        def process_well(well_path):
            group_handler = self._group_handler.derive_handler(well_path)
            well = OmeZarrWell(group_handler)
            return well_path, well

        wells, tasks = {}, []
        for well_path in self.wells_paths():
            task = asyncio.to_thread(process_well, well_path)
            tasks.append(task)

        results = await asyncio.gather(*tasks)
        for well_path, well in results:
            wells[well_path] = well

        self._group_handler.add_to_cache("wells", wells)
        return wells

    def get_wells(self) -> dict[str, OmeZarrWell]:
        """Get all wells in the plate.

        Returns:
            dict[str, OmeZarrWell]: A dictionary of wells, where the key is the well
                path and the value is the well object.
        """
        wells = self._group_handler.get_from_cache("wells")
        if wells is not None:
            assert isinstance(wells, dict)
            return wells

        def process_well(well_path):
            group_handler = self._group_handler.derive_handler(well_path)
            well = OmeZarrWell(group_handler)
            return well_path, well

        wells = {}
        for well_path in self.wells_paths():
            _, well = process_well(well_path)
            wells[well_path] = well

        self._group_handler.add_to_cache("wells", wells)
        return wells

    async def get_images_async(
        self, acquisition: int | None = None
    ) -> dict[str, OmeZarrContainer]:
        """Get all images in the plate asynchronously.

        This method processes images in parallel for improved performance
        when working with a large number of images.

        Args:
            acquisition: The acquisition id to filter the images.

        Returns:
            dict[str, OmeZarrContainer]: A dictionary of images, where the key is the
                image path and the value is the image object.
        """
        images = self._group_handler.get_from_cache("images")
        if images is not None:
            assert isinstance(images, dict)
            return images

        paths = await self.images_paths_async(acquisition=acquisition)

        def process_image(image_path):
            """Process a single image and return the image path and image object."""
            img_group_handler = self._group_handler.derive_handler(image_path)
            image = OmeZarrContainer(img_group_handler)
            return image_path, image

        images, tasks = {}, []
        for image_path in paths:
            task = asyncio.to_thread(process_image, image_path)
            tasks.append(task)

        results = await asyncio.gather(*tasks)

        for image_path, image in results:
            images[image_path] = image

        self._group_handler.add_to_cache("images", images)
        return images

    def get_images(self, acquisition: int | None = None) -> dict[str, OmeZarrContainer]:
        """Get all images in the plate.

        Args:
            acquisition: The acquisition id to filter the images.
        """
        images = self._group_handler.get_from_cache("images")
        if images is not None:
            assert isinstance(images, dict)
            return images
        paths = self.images_paths(acquisition=acquisition)

        def process_image(image_path):
            """Process a single image and return the image path and image object."""
            img_group_handler = self._group_handler.derive_handler(image_path)
            image = OmeZarrContainer(img_group_handler)
            return image_path, image

        images = {}
        for image_path in paths:
            _, image = process_image(image_path)
            images[image_path] = image

        self._group_handler.add_to_cache("images", images)
        return images

    def get_image(
        self, row: str, column: int | str, image_path: str
    ) -> OmeZarrContainer:
        """Get an image from the plate.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.
            image_path (str): The path of the image.

        Returns:
            OmeZarrContainer: The image.
        """
        image_path = self._image_path(row=row, column=column, path=image_path)
        group_handler = self._group_handler.derive_handler(image_path)
        return OmeZarrContainer(group_handler)

    def get_image_store(
        self, row: str, column: int | str, image_path: str
    ) -> StoreOrGroup:
        """Get the image store from the plate.

        Args:
            row (str): The row of the well.
            column (int | str): The column of the well.
            image_path (str): The path of the image.
        """
        well = self.get_well(row=row, column=column)
        return well.get_image_store(image_path=image_path)

    def get_well_images(
        self, row: str, column: str | int, acquisition: int | None = None
    ) -> dict[str, OmeZarrContainer]:
        """Get all images in a well.

        Args:
            row: The row of the well.
            column: The column of the well.
            acquisition: The acquisition id to filter the images.
        """
        images = {}
        for image_paths in self.well_images_paths(
            row=row, column=column, acquisition=acquisition
        ):
            group_handler = self._group_handler.derive_handler(image_paths)
            images[image_paths] = OmeZarrContainer(group_handler)
        return images

    def _add_image(
        self,
        row: str,
        column: int | str,
        image_path: str | None = None,
        acquisition_id: int | None = None,
        acquisition_name: str | None = None,
        atomic: bool = False,
    ) -> str:
        """Add an image to an ome-zarr plate."""
        if image_path is not None:
            image_path = path_in_well_validation(path=image_path)

        if atomic:
            plate_lock = self._group_handler.lock
        else:
            plate_lock = MockLock()

        with plate_lock:
            meta = self.meta
            meta = meta.add_well(row=row, column=column)
            if acquisition_id is not None:
                meta = meta.add_acquisition(
                    acquisition_id=acquisition_id, acquisition_name=acquisition_name
                )
            self.meta_handler.write_meta(meta)
            self.meta_handler._group_handler.clean_cache()

        well_path = self.meta.get_well_path(row=row, column=column)
        group_handler = self._group_handler.derive_handler(well_path)

        if atomic:
            well_lock = group_handler.lock
        else:
            well_lock = MockLock()

        with well_lock:
            attrs = group_handler.load_attrs()
            if len(attrs) == 0:
                # Initialize the well metadata
                # if the group is empty
                well_meta = NgioWellMeta.default_init()
                version = self.meta.plate.version
                version = version if version is not None else "0.4"
                meta_handler = get_well_meta_handler(group_handler, version=version)
            else:
                meta_handler = find_well_meta_handler(group_handler)
                well_meta = meta_handler.meta

            group_handler = self._group_handler.derive_handler(well_path)

            if image_path is not None:
                well_meta = well_meta.add_image(
                    path=image_path, acquisition=acquisition_id, strict=False
                )
            meta_handler.write_meta(well_meta)
            meta_handler._group_handler.clean_cache()

        if image_path is not None:
            return f"{well_path}/{image_path}"
        return well_path

    def atomic_add_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        acquisition_id: int | None = None,
        acquisition_name: str | None = None,
    ) -> str:
        """Parallel safe version of add_image."""
        if image_path is None:
            raise ValueError(
                "Image path cannot be None for atomic add_image. "
                "If your intent is to add a well, use add_well instead."
            )
        path = self._add_image(
            row=row,
            column=column,
            image_path=image_path,
            acquisition_id=acquisition_id,
            acquisition_name=acquisition_name,
            atomic=True,
        )
        return path

    def add_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        acquisition_id: int | None = None,
        acquisition_name: str | None = None,
    ) -> str:
        """Add an image to an ome-zarr plate."""
        if image_path is None:
            raise ValueError(
                "Image path cannot be None for atomic add_image. "
                "If your intent is to add a well, use add_well instead."
            )
        path = self._add_image(
            row=row,
            column=column,
            image_path=image_path,
            acquisition_id=acquisition_id,
            acquisition_name=acquisition_name,
            atomic=False,
        )
        return path

    def add_well(
        self,
        row: str,
        column: int | str,
    ) -> OmeZarrWell:
        """Add a well to an ome-zarr plate."""
        _ = self._add_image(
            row=row,
            column=column,
            image_path=None,
            acquisition_id=None,
            acquisition_name=None,
            atomic=False,
        )
        return self.get_well(row=row, column=column)

    def add_column(
        self,
        column: int | str,
    ) -> "OmeZarrPlate":
        """Add a column to an ome-zarr plate."""
        meta, _ = self.meta.add_column(column)
        self.meta_handler.write_meta(meta)
        self.meta_handler._group_handler.clean_cache()
        return self

    def add_row(
        self,
        row: str,
    ) -> "OmeZarrPlate":
        """Add a row to an ome-zarr plate."""
        meta, _ = self.meta.add_row(row)
        self.meta_handler.write_meta(meta)
        self.meta_handler._group_handler.clean_cache()
        return self

    def add_acquisition(
        self,
        acquisition_id: int,
        acquisition_name: str,
    ) -> "OmeZarrPlate":
        """Add an acquisition to an ome-zarr plate.

        Be aware that this is not a parallel safe operation.

        Args:
            acquisition_id (int): The acquisition id.
            acquisition_name (str): The acquisition name.
        """
        meta = self.meta.add_acquisition(
            acquisition_id=acquisition_id, acquisition_name=acquisition_name
        )
        self.meta_handler.write_meta(meta)
        self.meta_handler._group_handler.clean_cache()
        return self

    def _remove_well(
        self,
        row: str,
        column: int | str,
        atomic: bool = False,
    ):
        """Remove a well from an ome-zarr plate."""
        if atomic:
            plate_lock = self._group_handler.lock
        else:
            plate_lock = MockLock()

        with plate_lock:
            meta = self.meta
            meta = meta.remove_well(row, column)
            self.meta_handler.write_meta(meta)
            self.meta_handler._group_handler.clean_cache()

    def _remove_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
        atomic: bool = False,
    ):
        """Remove an image from an ome-zarr plate."""
        well = self.get_well(row, column)

        if atomic:
            well_lock = well.meta_handler._group_handler.lock
        else:
            well_lock = MockLock()

        with well_lock:
            well_meta = well.meta
            well_meta = well_meta.remove_image(path=image_path)
            well.meta_handler.write_meta(well_meta)
            well.meta_handler._group_handler.clean_cache()
            if len(well_meta.paths()) == 0:
                self._remove_well(row, column, atomic=atomic)

    def atomic_remove_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
    ):
        """Parallel safe version of remove_image."""
        return self._remove_image(
            row=row,
            column=column,
            image_path=image_path,
            atomic=True,
        )

    def remove_image(
        self,
        row: str,
        column: int | str,
        image_path: str,
    ):
        """Remove an image from an ome-zarr plate."""
        return self._remove_image(
            row=row,
            column=column,
            image_path=image_path,
            atomic=False,
        )

    def derive_plate(
        self,
        store: StoreOrGroup,
        plate_name: str | None = None,
        version: NgffVersions = "0.4",
        keep_acquisitions: bool = False,
        cache: bool = False,
        overwrite: bool = False,
        parallel_safe: bool = True,
    ) -> "OmeZarrPlate":
        """Derive a new OME-Zarr plate from an existing one.

        Args:
            store (StoreOrGroup): The Zarr store or group that stores the plate.
            plate_name (str | None): The name of the new plate.
            version (NgffVersion): The version of the new plate.
            keep_acquisitions (bool): Whether to keep the acquisitions in the new plate.
            cache (bool): Whether to use a cache for the zarr group metadata.
            overwrite (bool): Whether to overwrite the existing plate.
            parallel_safe (bool): Whether the group handler is parallel safe.
        """
        return derive_ome_zarr_plate(
            ome_zarr_plate=self,
            store=store,
            plate_name=plate_name,
            version=version,
            keep_acquisitions=keep_acquisitions,
            cache=cache,
            overwrite=overwrite,
            parallel_safe=parallel_safe,
        )

    def _get_tables_container(self) -> TablesContainer | None:
        """Return the tables container."""
        if self._tables_container is None:
            _tables_container = _default_table_container(self._group_handler)
            if _tables_container is None:
                return None
            self._tables_container = _tables_container
        return self._tables_container

    @property
    def tables_container(self) -> TablesContainer:
        """Return the tables container."""
        _table_container = self._get_tables_container()
        if _table_container is None:
            raise NgioValueError(
                "No tables container found. Please add a tables container to the plate."
            )
        return _table_container

    def list_tables(self, filter_types: TypedTable | str | None = None) -> list[str]:
        """List all tables in the image."""
        return self.tables_container.list(filter_types=filter_types)

    def list_roi_tables(self) -> list[str]:
        """List all ROI tables in the image."""
        masking_roi = self.tables_container.list(
            filter_types="masking_roi_table",
        )
        roi = self.tables_container.list(
            filter_types="roi_table",
        )
        return masking_roi + roi

    def get_roi_table(self, name: str) -> RoiTable:
        """Get a ROI table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, RoiTable):
            raise NgioValueError(f"Table {name} is not a ROI table. Got {type(table)}")
        return table

    def get_masking_roi_table(self, name: str) -> MaskingRoiTable:
        """Get a masking ROI table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, MaskingRoiTable):
            raise NgioValueError(
                f"Table {name} is not a masking ROI table. Got {type(table)}"
            )
        return table

    def get_feature_table(self, name: str) -> FeatureTable:
        """Get a feature table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, FeatureTable):
            raise NgioValueError(
                f"Table {name} is not a feature table. Got {type(table)}"
            )
        return table

    def get_generic_roi_table(self, name: str) -> GenericRoiTable:
        """Get a generic ROI table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, GenericRoiTable):
            raise NgioValueError(
                f"Table {name} is not a generic ROI table. Got {type(table)}"
            )
        return table

    def get_condition_table(self, name: str) -> ConditionTable:
        """Get a condition table from the image.

        Args:
            name (str): The name of the table.
        """
        table = self.tables_container.get(name=name, strict=True)
        if not isinstance(table, ConditionTable):
            raise NgioValueError(
                f"Table {name} is not a condition table. Got {type(table)}"
            )
        return table

    def get_table(self, name: str, check_type: TypedTable | None = None) -> Table:
        """Get a table from the image.

        Args:
            name (str): The name of the table.
            check_type (TypedTable | None): Deprecated. Please use
                'get_table_as' instead, or one of the type specific
                get_*table() methods.

        """
        if check_type is not None:
            warnings.warn(
                "The 'check_type' argument is deprecated, and will be removed in "
                "ngio=0.3. Use 'get_table_as' instead or one of the "
                "type specific get_*table() methods.",
                DeprecationWarning,
                stacklevel=2,
            )
        return self.tables_container.get(name=name, strict=False)

    def get_table_as(
        self,
        name: str,
        table_cls: type[TableType],
        backend: TableBackend | None = None,
    ) -> TableType:
        """Get a table from the image as a specific type.

        Args:
            name (str): The name of the table.
            table_cls (type[TableType]): The type of the table.
            backend (TableBackend | None): The backend to use. If None,
                the default backend is used.
        """
        return self.tables_container.get_as(
            name=name,
            table_cls=table_cls,
            backend=backend,
        )

    def add_table(
        self,
        name: str,
        table: Table,
        backend: TableBackend = DefaultTableBackend,
        overwrite: bool = False,
    ) -> None:
        """Add a table to the image."""
        self.tables_container.add(
            name=name, table=table, backend=backend, overwrite=overwrite
        )

    def list_image_tables(
        self,
        acquisition: int | None = None,
        filter_types: str | None = None,
        mode: Literal["common", "all"] = "common",
    ) -> list[str]:
        """List all image tables in the image.

        Args:
            acquisition (int | None): The acquisition id to filter the images.
            filter_types (str | None): The type of tables to filter. If None,
                return all tables. Defaults to None.
            mode (Literal["common", "all"]): The mode to use for listing the tables.
                If 'common', return only common tables between all images.
                If 'all', return all tables. Defaults to 'common'.
        """
        images = tuple(self.get_images(acquisition=acquisition).values())
        return list_image_tables(
            images=images,
            filter_types=filter_types,
            mode=mode,
        )

    async def list_image_tables_async(
        self,
        acquisition: int | None = None,
        filter_types: str | None = None,
        mode: Literal["common", "all"] = "common",
    ) -> list[str]:
        """List all image tables in the image asynchronously.

        Args:
            acquisition (int | None): The acquisition id to filter the images.
            filter_types (str | None): The type of tables to filter. If None,
                return all tables. Defaults to None.
            mode (Literal["common", "all"]): The mode to use for listing the tables.
                If 'common', return only common tables between all images.
                If 'all', return all tables. Defaults to 'common'.
        """
        images = await self.get_images_async(acquisition=acquisition)
        images = tuple(images.values())
        return await list_image_tables_async(
            images=images,
            filter_types=filter_types,
            mode=mode,
        )

    def concatenate_image_tables(
        self,
        name: str,
        acquisition: int | None = None,
        strict: bool = True,
        index_key: str | None = None,
        mode: Literal["eager", "lazy"] = "eager",
    ) -> Table:
        """Concatenate tables from all images in the plate.

        Args:
            name: The name of the table to concatenate.
            index_key: The key to use for the index of the concatenated table.
            acquisition: The acquisition id to filter the images.
            strict: If True, raise an error if the table is not found in the image.
            index_key: If a string is provided, a new index column will be created
                new_index_pattern = {row}_{column}_{path_in_well}_{label}
            mode: The mode to use for concatenation. Can be 'eager' or 'lazy'.
                if 'eager', the table will be loaded into memory.
                if 'lazy', the table will be loaded as a lazy frame.
        """
        images = self.get_images(acquisition=acquisition)
        extras = _build_extras(tuple(images.keys()))
        return concatenate_image_tables(
            images=tuple(images.values()),
            extras=extras,
            name=name,
            index_key=index_key,
            strict=strict,
            mode=mode,
        )

    def concatenate_image_tables_as(
        self,
        name: str,
        table_cls: type[TableType],
        acquisition: int | None = None,
        index_key: str | None = None,
        strict: bool = True,
        mode: Literal["eager", "lazy"] = "eager",
    ) -> TableType:
        """Concatenate tables from all images in the plate as a specific type.

        Args:
            name: The name of the table to concatenate.
            table_cls: The type of the table to concatenate.
            index_key: The key to use for the index of the concatenated table.
            acquisition: The acquisition id to filter the images.
            index_key: If a string is provided, a new index column will be created
                new_index_pattern = {row}_{column}_{path_in_well}_{label}
            strict: If True, raise an error if the table is not found in the image.
            mode: The mode to use for concatenation. Can be 'eager' or 'lazy'.
                if 'eager', the table will be loaded into memory.
                if 'lazy', the table will be loaded as a lazy frame.
        """
        images = self.get_images(acquisition=acquisition)
        extras = _build_extras(tuple(images.keys()))
        return concatenate_image_tables_as(
            images=tuple(images.values()),
            extras=extras,
            name=name,
            table_cls=table_cls,
            index_key=index_key,
            strict=strict,
            mode=mode,
        )

    async def concatenate_image_tables_async(
        self,
        name: str,
        acquisition: int | None = None,
        index_key: str | None = None,
        strict: bool = True,
        mode: Literal["eager", "lazy"] = "eager",
    ) -> Table:
        """Concatenate tables from all images in the plate asynchronously.

        Args:
            name: The name of the table to concatenate.
            index_key: The key to use for the index of the concatenated table.
            acquisition: The acquisition id to filter the images.
            index_key: If a string is provided, a new index column will be created
                new_index_pattern = {row}_{column}_{path_in_well}_{label}
            strict: If True, raise an error if the table is not found in the image.
            mode: The mode to use for concatenation. Can be 'eager' or 'lazy'.
                if 'eager', the table will be loaded into memory.
                if 'lazy', the table will be loaded as a lazy frame.
        """
        images = await self.get_images_async(acquisition=acquisition)
        extras = _build_extras(tuple(images.keys()))
        return await concatenate_image_tables_async(
            images=tuple(images.values()),
            extras=extras,
            name=name,
            index_key=index_key,
            strict=strict,
            mode=mode,
        )

    async def concatenate_image_tables_as_async(
        self,
        name: str,
        table_cls: type[TableType],
        acquisition: int | None = None,
        index_key: str | None = None,
        strict: bool = True,
        mode: Literal["eager", "lazy"] = "eager",
    ) -> TableType:
        """Concatenate tables from all images in the plate as a specific type.

        Args:
            name: The name of the table to concatenate.
            table_cls: The type of the table to concatenate.
            index_key: The key to use for the index of the concatenated table.
            acquisition: The acquisition id to filter the images.
            index_key: If a string is provided, a new index column will be created
                new_index_pattern = {row}_{column}_{path_in_well}_{label}
            strict: If True, raise an error if the table is not found in the image.
            mode: The mode to use for concatenation. Can be 'eager' or 'lazy'.
                if 'eager', the table will be loaded into memory.
                if 'lazy', the table will be loaded as a lazy frame.
        """
        images = await self.get_images_async(acquisition=acquisition)
        extras = _build_extras(tuple(images.keys()))
        return await concatenate_image_tables_as_async(
            images=tuple(images.values()),
            extras=extras,
            name=name,
            table_cls=table_cls,
            index_key=index_key,
            strict=strict,
            mode=mode,
        )


def open_ome_zarr_plate(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
    parallel_safe: bool = True,
) -> OmeZarrPlate:
    """Open an OME-Zarr plate.

    Args:
        store (StoreOrGroup): The Zarr store or group that stores the plate.
        cache (bool): Whether to use a cache for the zarr group metadata.
        mode (AccessModeLiteral): The
            access mode for the image. Defaults to "r+".
        parallel_safe (bool): Whether the group handler is parallel safe.
    """
    group_handler = ZarrGroupHandler(
        store=store, cache=cache, mode=mode, parallel_safe=parallel_safe
    )
    return OmeZarrPlate(group_handler)


def _create_empty_plate_from_meta(
    store: StoreOrGroup,
    meta: NgioPlateMeta,
    version: NgffVersions = "0.4",
    overwrite: bool = False,
) -> ZarrGroupHandler:
    """Create an empty OME-Zarr plate from metadata."""
    mode = "w" if overwrite else "w-"
    group_handler = ZarrGroupHandler(
        store=store, cache=True, mode=mode, parallel_safe=False
    )
    meta_handler = get_plate_meta_handler(group_handler, version=version)
    meta_handler.write_meta(meta)
    return group_handler


def create_empty_plate(
    store: StoreOrGroup,
    name: str,
    images: list[ImageInWellPath] | None = None,
    version: NgffVersions = "0.4",
    cache: bool = False,
    overwrite: bool = False,
    parallel_safe: bool = True,
) -> OmeZarrPlate:
    """Initialize and create an empty OME-Zarr plate."""
    plate_meta = NgioPlateMeta.default_init(
        name=name,
        version=version,
    )
    group_handler = _create_empty_plate_from_meta(
        store=store,
        meta=plate_meta,
        version=version,
        overwrite=overwrite,
    )

    if images is not None:
        plate = OmeZarrPlate(group_handler)
        for image in images:
            plate.add_image(
                row=image.row,
                column=image.column,
                image_path=image.path,
                acquisition_id=image.acquisition_id,
                acquisition_name=image.acquisition_name,
            )
    return open_ome_zarr_plate(
        store=store,
        cache=cache,
        mode="r+",
        parallel_safe=parallel_safe,
    )


def derive_ome_zarr_plate(
    ome_zarr_plate: OmeZarrPlate,
    store: StoreOrGroup,
    plate_name: str | None = None,
    version: NgffVersions = "0.4",
    keep_acquisitions: bool = False,
    cache: bool = False,
    overwrite: bool = False,
    parallel_safe: bool = True,
) -> OmeZarrPlate:
    """Derive a new OME-Zarr plate from an existing one.

    Args:
        ome_zarr_plate (OmeZarrPlate): The existing OME-Zarr plate.
        store (StoreOrGroup): The Zarr store or group that stores the plate.
        plate_name (str | None): The name of the new plate.
        version (NgffVersion): The version of the new plate.
        keep_acquisitions (bool): Whether to keep the acquisitions in the new plate.
        cache (bool): Whether to use a cache for the zarr group metadata.
        overwrite (bool): Whether to overwrite the existing plate.
        parallel_safe (bool): Whether the group handler is parallel safe.
    """
    if plate_name is None:
        plate_name = ome_zarr_plate.meta.plate.name

    new_meta = ome_zarr_plate.meta.derive(
        name=plate_name,
        version=version,
        keep_acquisitions=keep_acquisitions,
    )
    _ = _create_empty_plate_from_meta(
        store=store,
        meta=new_meta,
        overwrite=overwrite,
        version=version,
    )
    return open_ome_zarr_plate(
        store=store,
        cache=cache,
        mode="r+",
        parallel_safe=parallel_safe,
    )


def open_ome_zarr_well(
    store: StoreOrGroup,
    cache: bool = False,
    mode: AccessModeLiteral = "r+",
    parallel_safe: bool = True,
) -> OmeZarrWell:
    """Open an OME-Zarr well.

    Args:
        store (StoreOrGroup): The Zarr store or group that stores the plate.
        cache (bool): Whether to use a cache for the zarr group metadata.
        mode (AccessModeLiteral): The access mode for the image. Defaults to "r+".
        parallel_safe (bool): Whether the group handler is parallel safe.
    """
    group_handler = ZarrGroupHandler(
        store=store, cache=cache, mode=mode, parallel_safe=parallel_safe
    )
    return OmeZarrWell(group_handler)


def create_empty_well(
    store: StoreOrGroup,
    version: NgffVersions = "0.4",
    cache: bool = False,
    overwrite: bool = False,
    parallel_safe: bool = True,
) -> OmeZarrWell:
    """Create an empty OME-Zarr well.

    Args:
        store (StoreOrGroup): The Zarr store or group that stores the well.
        version (NgffVersion): The version of the new well.
        cache (bool): Whether to use a cache for the zarr group metadata.
        overwrite (bool): Whether to overwrite the existing well.
        parallel_safe (bool): Whether the group handler is parallel safe.
    """
    group_handler = ZarrGroupHandler(
        store=store, cache=True, mode="w" if overwrite else "w-", parallel_safe=False
    )
    meta_handler = get_well_meta_handler(group_handler, version=version)
    meta = NgioWellMeta.default_init()
    meta_handler.write_meta(meta)

    return open_ome_zarr_well(
        store=store,
        cache=cache,
        mode="r+",
        parallel_safe=parallel_safe,
    )



import logging
import time
from functools import cache

from ngio.utils._errors import NgioValueError

# Configure the logger
ngio_logger = logging.getLogger("NgioLogger")
ngio_logger.setLevel(logging.ERROR)

# Set up a console handler with a custom format
console_handler = logging.StreamHandler()
formatter = logging.Formatter(
    "%(asctime)s - %(levelname)s - %(name)s - "
    "[%(module)s.%(funcName)s:%(lineno)d]: %(message)s"
)
console_handler.setFormatter(formatter)

# Add the handler to the logger
ngio_logger.addHandler(console_handler)


def set_logger_level(level: str) -> None:
    """Set the logger level.

    Args:
        level: The level to set the logger to.
            Must be one of "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL".
    """
    if level not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        raise NgioValueError(f"Invalid log level: {level}")

    ngio_logger.setLevel(level)


@cache
def _warn(message: str, ttl_hash: int) -> None:
    """Log a warning message with a time-to-live (TTL) hash."""
    ngio_logger.warning(message, stacklevel=3)


def ngio_warn(message: str, cooldown: int = 2) -> None:
    """Log a warning message.

    Args:
        message: The warning message to log.
        cooldown: The cooldown period in seconds to avoid repeated logging.
    """
    ttl_hash = time.time() // cooldown
    _warn(message, ttl_hash)
